# -*- coding: utf-8 -*-
"""
Collection of raw lexer test cases and class constructor.
"""

from __future__ import unicode_literals

import textwrap

swapquotes = {
    39: 34, 34: 39,
    # note the follow are interim error messages
    96: 39,
}

# The structure and some test cases are taken
# from https://bitbucket.org/ned/jslex
es5_cases = [
    (
        # Identifiers
        'identifiers_ascii',
        ('i my_variable_name c17 _dummy $str $ _ CamelCase class2type',
         ['ID i', 'ID my_variable_name', 'ID c17', 'ID _dummy',
          'ID $str', 'ID $', 'ID _', 'ID CamelCase', 'ID class2type']
         ),
    ), (
        'identifiers_unicode',
        (u'\u03c0 \u03c0_tail var\ua67c',
         [u'ID \u03c0', u'ID \u03c0_tail', u'ID var\ua67c']),
    ), (
        # https://github.com/rspivak/slimit/issues/2
        'slimit_issue_2',
        ('nullify truelie falsepositive',
         ['ID nullify', 'ID truelie', 'ID falsepositive']),

    ), (
        'keywords_break',
        ('break Break BREAK', ['BREAK break', 'ID Break', 'ID BREAK']),
    ), (
        # Literals
        'literals',
        ('null true false Null True False',
         ['NULL null', 'TRUE true', 'FALSE false',
          'ID Null', 'ID True', 'ID False']
         ),
    ), (
        # Punctuators
        'punctuators_simple',
        ('a /= b', ['ID a', 'DIVEQUAL /=', 'ID b']),
    ), (
        'punctuators_various_equality',
        (('= == != === !== < > <= >= || && ++ -- << >> '
          '>>> += -= *= <<= >>= >>>= &= %= ^= |='),
         ['EQ =', 'EQEQ ==', 'NE !=', 'STREQ ===', 'STRNEQ !==', 'LT <',
          'GT >', 'LE <=', 'GE >=', 'OR ||', 'AND &&', 'PLUSPLUS ++',
          'MINUSMINUS --', 'LSHIFT <<', 'RSHIFT >>', 'URSHIFT >>>',
          'PLUSEQUAL +=', 'MINUSEQUAL -=', 'MULTEQUAL *=', 'LSHIFTEQUAL <<=',
          'RSHIFTEQUAL >>=', 'URSHIFTEQUAL >>>=', 'ANDEQUAL &=', 'MODEQUAL %=',
          'XOREQUAL ^=', 'OREQUAL |=',
          ]
         ),
    ), (
        'punctuators_various_others',
        ('. , ; : + - * % & | ^ ~ ? ! ( ) { } [ ]',
         ['PERIOD .', 'COMMA ,', 'SEMI ;', 'COLON :', 'PLUS +', 'MINUS -',
          'MULT *', 'MOD %', 'BAND &', 'BOR |', 'BXOR ^', 'BNOT ~',
          'CONDOP ?', 'NOT !', 'LPAREN (', 'RPAREN )', 'LBRACE {', 'RBRACE }',
          'LBRACKET [', 'RBRACKET ]']
         ),
    ), (
        'division_simple',
        ('a / b', ['ID a', 'DIV /', 'ID b']),
    ), (
        'numbers',
        (('3 3.3 0 0. 0.0 0.001 010 3.e2 3.e-2 3.e+2 3E2 3E+2 3E-2 '
          '0.5e2 0.5e+2 0.5e-2 33 128.15 0x001 0X12ABCDEF 0xabcdef'),
         ['NUMBER 3', 'NUMBER 3.3', 'NUMBER 0', 'NUMBER 0.', 'NUMBER 0.0',
          'NUMBER 0.001', 'NUMBER 010', 'NUMBER 3.e2', 'NUMBER 3.e-2',
          'NUMBER 3.e+2', 'NUMBER 3E2', 'NUMBER 3E+2', 'NUMBER 3E-2',
          'NUMBER 0.5e2', 'NUMBER 0.5e+2', 'NUMBER 0.5e-2', 'NUMBER 33',
          'NUMBER 128.15', 'NUMBER 0x001', 'NUMBER 0X12ABCDEF',
          'NUMBER 0xabcdef']
         ),
    ), (
        'strings_simple_quote',
        (""" '"' """, ["""STRING '"'"""]),
    ), (
        'strings_escape_quote_tab',
        (r'''"foo" 'foo' "x\";" 'x\';' "foo\tbar"''',
         ['STRING "foo"', """STRING 'foo'""", r'STRING "x\";"',
          r"STRING 'x\';'", r'STRING "foo\tbar"']
         ),
    ), (
        'strings_escape_ascii',
        (r"""'\x55' "\x12ABCDEF" '!@#$%^&*()_+{}[]\";?'""",
         [r"STRING '\x55'", r'STRING "\x12ABCDEF"',
          r"STRING '!@#$%^&*()_+{}[]\";?'"]
         ),
    ), (
        'strings_escape_unicode',
        (r"""'\u0001' "\uFCEF" 'a\\\b\n'""",
         [r"STRING '\u0001'", r'STRING "\uFCEF"', r"STRING 'a\\\b\n'"]
         ),
    ), (
        'strings_unicode',
        (u'"тест строки\\""', [u'STRING "тест строки\\""']),
    ), (
        'strings_escape_octal',
        (r"""'\251'""", [r"""STRING '\251'"""]),
    ), (
        # Bug - https://github.com/rspivak/slimit/issues/5
        'slimit_issue_5',
        (r"var tagRegExp = new RegExp('<(\/*)(FooBar)', 'gi');",
         ['VAR var', 'ID tagRegExp', 'EQ =',
          'NEW new', 'ID RegExp', 'LPAREN (',
          r"STRING '<(\/*)(FooBar)'", 'COMMA ,', "STRING 'gi'",
          'RPAREN )', 'SEMI ;']),
    ), (
        # same as above but inside double quotes
        'slimit_issue_5_double_quote',
        (r'"<(\/*)(FooBar)"', [r'STRING "<(\/*)(FooBar)"']),
    ), (
        # multiline string (string written across multiple lines
        # of code) https://github.com/rspivak/slimit/issues/24
        'slimit_issue_24_multi_line_code_double',
        ("var a = 'hello \\\n world'",
         ['VAR var', 'ID a', 'EQ =', "STRING 'hello \\\n world'"]),
    ), (
        'slimit_issue_24_multi_line_code_single',
        ('var a = "hello \\\r world"',
         ['VAR var', 'ID a', 'EQ =', 'STRING "hello \\\r world"']),
    ), (
        # regex
        'regex_1',
        (r'a=/a*/,1', ['ID a', 'EQ =', 'REGEX /a*/', 'COMMA ,', 'NUMBER 1']),
    ), (
        'regex_2',
        (r'a=/a*[^/]+/,1',
         ['ID a', 'EQ =', 'REGEX /a*[^/]+/', 'COMMA ,', 'NUMBER 1']
         ),
    ), (
        'regex_3',
        (r'a=/a*\[^/,1',
         ['ID a', 'EQ =', r'REGEX /a*\[^/', 'COMMA ,', 'NUMBER 1']
         ),
    ), (
        'regex_4',
        (r'a=/\//,1', ['ID a', 'EQ =', r'REGEX /\//', 'COMMA ,', 'NUMBER 1']),
    ), (
        # not a regex, just a division
        # https://github.com/rspivak/slimit/issues/6
        'slimit_issue_6_not_regex_but_division',
        (r'x = this / y;',
         ['ID x', 'EQ =', 'THIS this', r'DIV /', r'ID y', r'SEMI ;']),
    ), (
        'regex_mozilla_example_1',
        # next two are from
        # http://www.mozilla.org/js/language/js20-2002-04/rationale/syntax.html#regular-expressions
        ('for (var x = a in foo && "</x>" || mot ? z:/x:3;x<5;y</g/i) '
         '{xyz(x++);}',
         ["FOR for", "LPAREN (", "VAR var", "ID x", "EQ =", "ID a", "IN in",
          "ID foo", "AND &&", 'STRING "</x>"', "OR ||", "ID mot", "CONDOP ?",
          "ID z", "COLON :", "REGEX /x:3;x<5;y</g", "DIV /", "ID i",
          "RPAREN )", "LBRACE {", "ID xyz", "LPAREN (", "ID x", "PLUSPLUS ++",
          "RPAREN )", "SEMI ;", "RBRACE }"]
         ),
    ), (
        'regex_mozilla_example_2',
        ('for (var x = a in foo && "</x>" || mot ? z/x:3;x<5;y</g/i) '
         '{xyz(x++);}',
         ["FOR for", "LPAREN (", "VAR var", "ID x", "EQ =", "ID a", "IN in",
          "ID foo", "AND &&", 'STRING "</x>"', "OR ||", "ID mot", "CONDOP ?",
          "ID z", "DIV /", "ID x", "COLON :", "NUMBER 3", "SEMI ;", "ID x",
          "LT <", "NUMBER 5", "SEMI ;", "ID y", "LT <", "REGEX /g/i",
          "RPAREN )", "LBRACE {", "ID xyz", "LPAREN (", "ID x", "PLUSPLUS ++",
          "RPAREN )", "SEMI ;", "RBRACE }"]
         ),

    ), (
        'regex_illegal_1',
        # Various "illegal" regexes that are valid according to the std.
        (r"""/????/, /++++/, /[----]/ """,
         ['REGEX /????/', 'COMMA ,',
          'REGEX /++++/', 'COMMA ,', 'REGEX /[----]/']
         ),

    ), (
        'regex_stress_test_1',
        # Stress cases from
        # http://stackoverflow.com/questions/5533925/
        # what-javascript-constructs-does-jslex-incorrectly-lex/5573409#5573409
        (r"""/\[/""", [r"""REGEX /\[/"""]),
    ), (
        'regex_stress_test_2',
        (r"""/[i]/""", [r"""REGEX /[i]/"""]),
    ), (
        'regex_stress_test_3',
        (r"""/[\]]/""", [r"""REGEX /[\]]/"""]),
    ), (
        'regex_stress_test_4',
        (r"""/a[\]]/""", [r"""REGEX /a[\]]/"""]),
    ), (
        'regex_stress_test_5',
        (r"""/a[\]]b/""", [r"""REGEX /a[\]]b/"""]),
    ), (
        'regex_stress_test_6',
        (r"""/[\]/]/gi""", [r"""REGEX /[\]/]/gi"""]),
    ), (
        'regex_stress_test_7',
        (r"""/\[[^\]]+\]/gi""", [r"""REGEX /\[[^\]]+\]/gi"""]),
    ), (
        'regex_stress_test_8',
        (r"""
         rexl.re = {
         NAME: /^(?!\d)(?:\w)+|^"(?:[^"]|"")+"/,
         UNQUOTED_LITERAL: /^@(?:(?!\d)(?:\w|\:)+|^"(?:[^"]|"")+")\[[^\]]+\]/,
         QUOTED_LITERAL: /^'(?:[^']|'')*'/,
         NUMERIC_LITERAL: /^[0-9]+(?:\.[0-9]*(?:[eE][-+][0-9]+)?)?/,
         SYMBOL: /^(?:==|=|<>|<=|<|>=|>|!~~|!~|~~|~|!==|!=|!~=|!~|!|&|\||\.|\:|,|\(|\)|\[|\]|\{|\}|\?|\:|;|@|\^|\/\+|\/|\*|\+|-)/
         };
         """, [
         "ID rexl", "PERIOD .", "ID re", "EQ =", "LBRACE {",
         "ID NAME", "COLON :",
         r"""REGEX /^(?!\d)(?:\w)+|^"(?:[^"]|"")+"/""", "COMMA ,",
         "ID UNQUOTED_LITERAL", "COLON :",
         r"""REGEX /^@(?:(?!\d)(?:\w|\:)+|^"(?:[^"]|"")+")\[[^\]]+\]/""",
         "COMMA ,", "ID QUOTED_LITERAL", "COLON :",
         r"""REGEX /^'(?:[^']|'')*'/""", "COMMA ,", "ID NUMERIC_LITERAL",
         "COLON :",
         r"""REGEX /^[0-9]+(?:\.[0-9]*(?:[eE][-+][0-9]+)?)?/""", "COMMA ,",
         "ID SYMBOL", "COLON :",
         r"""REGEX /^(?:==|=|<>|<=|<|>=|>|!~~|!~|~~|~|!==|!=|!~=|!~|!|&|\||\.|\:|,|\(|\)|\[|\]|\{|\}|\?|\:|;|@|\^|\/\+|\/|\*|\+|-)/""",
         "RBRACE }", "SEMI ;"]),
    ), (
        'regex_stress_test_9',
        (r"""
         rexl.re = {
         NAME: /^(?!\d)(?:\w)+|^"(?:[^"]|"")+"/,
         UNQUOTED_LITERAL: /^@(?:(?!\d)(?:\w|\:)+|^"(?:[^"]|"")+")\[[^\]]+\]/,
         QUOTED_LITERAL: /^'(?:[^']|'')*'/,
         NUMERIC_LITERAL: /^[0-9]+(?:\.[0-9]*(?:[eE][-+][0-9]+)?)?/,
         SYMBOL: /^(?:==|=|<>|<=|<|>=|>|!~~|!~|~~|~|!==|!=|!~=|!~|!|&|\||\.|\:|,|\(|\)|\[|\]|\{|\}|\?|\:|;|@|\^|\/\+|\/|\*|\+|-)/
         };
         str = '"';
         """, [
         "ID rexl", "PERIOD .", "ID re", "EQ =", "LBRACE {",
         "ID NAME", "COLON :", r"""REGEX /^(?!\d)(?:\w)+|^"(?:[^"]|"")+"/""",
         "COMMA ,", "ID UNQUOTED_LITERAL", "COLON :",
         r"""REGEX /^@(?:(?!\d)(?:\w|\:)+|^"(?:[^"]|"")+")\[[^\]]+\]/""",
         "COMMA ,", "ID QUOTED_LITERAL", "COLON :",
         r"""REGEX /^'(?:[^']|'')*'/""", "COMMA ,",
         "ID NUMERIC_LITERAL", "COLON :",
         r"""REGEX /^[0-9]+(?:\.[0-9]*(?:[eE][-+][0-9]+)?)?/""", "COMMA ,",
         "ID SYMBOL", "COLON :",
         r"""REGEX /^(?:==|=|<>|<=|<|>=|>|!~~|!~|~~|~|!==|!=|!~=|!~|!|&|\||\.|\:|,|\(|\)|\[|\]|\{|\}|\?|\:|;|@|\^|\/\+|\/|\*|\+|-)/""",
         "RBRACE }", "SEMI ;",
         "ID str", "EQ =", """STRING '"'""", "SEMI ;",
         ]),
    ), (
        'regex_stress_test_10',
        (r""" this._js = "e.str(\"" + this.value.replace(/\\/g, "\\\\").replace(/"/g, "\\\"") + "\")"; """,
         ["THIS this", "PERIOD .", "ID _js", "EQ =",
          r'''STRING "e.str(\""''', "PLUS +", "THIS this", "PERIOD .",
          "ID value", "PERIOD .", "ID replace", "LPAREN (", r"REGEX /\\/g",
          "COMMA ,", r'STRING "\\\\"', "RPAREN )", "PERIOD .", "ID replace",
          "LPAREN (", r'REGEX /"/g', "COMMA ,", r'STRING "\\\""', "RPAREN )",
          "PLUS +", r'STRING "\")"', "SEMI ;"]),
    ), (
        'regex_division_check',
        ('a = /a/ / /b/',
         ['ID a', 'EQ =', 'REGEX /a/', 'DIV /', 'REGEX /b/']),
    ), (
        'regex_after_plus_brace',
        ('+{}/a/g',
         ['PLUS +', 'LBRACE {', 'RBRACE }', 'DIV /', 'ID a', 'DIV /', 'ID g']),
        # The following pathological cases cannot be tested using the
        # lexer alone, as the rules can only be addressed in conjunction
        # with a parser
        #
        # 'regex_after_brace',
        # ('{}/a/g',
        #  ['LBRACE {', 'RBRACE }', 'REGEX /a/g']),
        # 'regex_after_if_brace',
        # ('if (a) { } /a/.test(a)',
        #  ['IF if', 'LPAREN (', 'ID a', 'RPAREN )', 'LBRACE {', 'RBRACE }',
        #   'REGEX /a/', "PERIOD .", "ID test", 'LPAREN (', 'ID a',
        #   'RPAREN )']),
    ), (
        'regex_case',
        ('switch(0){case /a/:}',
         ['SWITCH switch', 'LPAREN (', 'NUMBER 0', 'RPAREN )', 'LBRACE {',
          'CASE case', 'REGEX /a/', 'COLON :', 'RBRACE }']),
    ), (
        'div_after_valid_statement_function_call',
        ('if(){} f(a) / f(b)',
         ['IF if', 'LPAREN (', 'RPAREN )', 'LBRACE {', 'RBRACE }',
          'ID f', 'LPAREN (', 'ID a', 'RPAREN )', 'DIV /',
          'ID f', 'LPAREN (', 'ID b', 'RPAREN )']),
    ), (
        'for_regex_slimit_issue_54',
        ('for (;;) /r/;',
         ['FOR for', 'LPAREN (', 'SEMI ;', 'SEMI ;', 'RPAREN )',
          'REGEX /r/', 'SEMI ;']),
    ), (
        'for_regex_slimit_issue_54_not_break_division',
        ('for (;;) { x / y }',
         ['FOR for', 'LPAREN (', 'SEMI ;', 'SEMI ;', 'RPAREN )',
          'LBRACE {', 'ID x', 'DIV /', 'ID y', 'RBRACE }']),
    ), (
        'for_regex_slimit_issue_54_bracket_accessor_check',
        ('s = {a:1} + s[2] / 1',
         ['ID s', 'EQ =', 'LBRACE {', 'ID a', 'COLON :', 'NUMBER 1',
          'RBRACE }', 'PLUS +', 'ID s', 'LBRACKET [', 'NUMBER 2', 'RBRACKET ]',
          'DIV /', 'NUMBER 1'])
    ), (
        'for_regex_slimit_issue_54_function_parentheses_check',
        ('s = {a:1} + f(2) / 1',
         ['ID s', 'EQ =', 'LBRACE {', 'ID a', 'COLON :', 'NUMBER 1',
          'RBRACE }', 'PLUS +', 'ID f', 'LPAREN (', 'NUMBER 2', 'RPAREN )',
          'DIV /', 'NUMBER 1'])
    ), (
        'for_regex_slimit_issue_54_math_parentheses_check',
        ('s = {a:1} + (2) / 1',
         ['ID s', 'EQ =', 'LBRACE {', 'ID a', 'COLON :', 'NUMBER 1',
          'RBRACE }', 'PLUS +', 'LPAREN (', 'NUMBER 2', 'RPAREN )',
          'DIV /', 'NUMBER 1'])
    ), (
        'for_regex_slimit_issue_54_math_bracket_check',
        ('s = {a:1} + [2] / 1',
         ['ID s', 'EQ =', 'LBRACE {', 'ID a', 'COLON :', 'NUMBER 1',
          'RBRACE }', 'PLUS +', 'LBRACKET [', 'NUMBER 2', 'RBRACKET ]',
          'DIV /', 'NUMBER 1'])
    ), (
        'for_regex_slimit_issue_54_math_braces_check',
        ('s = {a:2} / 166 / 9',
         ['ID s', 'EQ =', 'LBRACE {', 'ID a', 'COLON :', 'NUMBER 2',
          'RBRACE }', 'DIV /', 'NUMBER 166', 'DIV /', 'NUMBER 9'])
    ), (
        'do_while_regex',
        ('do {} while (0) /s/',
         ['DO do', 'LBRACE {', 'RBRACE }', 'WHILE while', 'LPAREN (',
          'NUMBER 0', 'RPAREN )', 'REGEX /s/'])
    ), (
        'if_regex',
        ('if (thing) /s/',
         ['IF if', 'LPAREN (', 'ID thing', 'RPAREN )', 'REGEX /s/'])
    ), (
        'identifier_math',
        ('f (v) /s/g',
         ['ID f', 'LPAREN (', 'ID v', 'RPAREN )', 'DIV /', 'ID s', 'DIV /',
          'ID g'])
    ), (
        'section_7',
        ("a = b\n/hi/s",
         ['ID a', 'EQ =', 'ID b', 'DIV /', 'ID hi', 'DIV /', 'ID s'])
    ), (
        'section_7_extras',
        ("a = b\n\n\n/hi/s",
         ['ID a', 'EQ =', 'ID b', 'DIV /', 'ID hi', 'DIV /', 'ID s'])
    ), (
        'slimit_issue_39_and_57',
        (r"f(a, 'hi\01').split('\1').split('\0');",
         ['ID f', 'LPAREN (', 'ID a', 'COMMA ,', r"STRING 'hi\01'", 'RPAREN )',
          'PERIOD .', 'ID split', 'LPAREN (', r"STRING '\1'", 'RPAREN )',
          'PERIOD .', 'ID split', 'LPAREN (', r"STRING '\0'", 'RPAREN )',
          'SEMI ;'])
    ), (
        'section_7_8_4_string_literal_with_7_3_conformance',
        ("'<LF>\\\n<CR>\\\r<LS>\\\u2028<PS>\\\u2029<CR><LF>\\\r\n'",
         ["STRING '<LF>\\\n<CR>\\\r<LS>\\\u2028<PS>\\\u2029<CR><LF>\\\r\n'"])
    ), (
        # okay this is getting ridiculous how bad ECMA is.
        'section_7_comments',
        ("a = b\n/** **/\n\n/hi/s",
         ['ID a', 'EQ =', 'ID b', 'DIV /', 'ID hi', 'DIV /', 'ID s'])
    ),
]

# various string related syntax errors
es5_error_cases_str = [
    (
        'unterminated_string_eof',
        "var foo = 'test",
        'Unterminated string literal "\'test" at 1:11',
    ), (
        'naked_line_separator_in_string',
        "vaf foo = 'test\u2028foo'",
        'Unterminated string literal "\'test" at 1:11',
    ), (
        'naked_line_feed_in_string',
        "var foo = 'test\u2029foo'",
        'Unterminated string literal "\'test" at 1:11',
    ), (
        'naked_crnl_in_string',
        "var foo = 'test\r\nfoo'",
        'Unterminated string literal "\'test" at 1:11',
    ), (
        'naked_cr_in_string',
        "var foo = 'test\\\n\rfoo'",
        # FIXME Note that the \\ is double escaped
        'Unterminated string literal "\'test\\\\" at 1:11',
    ), (
        'invalid_hex_sequence',
        "var foo = 'fail\\x1'",
        # backticks are converted to single quotes
        "Invalid hexadecimal escape sequence `\\x1` at 1:16",
    ), (
        'invalid_unicode_sequence',
        "var foo = 'fail\\u12'",
        "Invalid unicode escape sequence `\\u12` at 1:16",
    ), (
        'invalid_hex_sequence_multiline',
        "var foo = 'foobar\\\r\nfail\\x1'",
        # backticks are converted to single quotes
        "Invalid hexadecimal escape sequence `\\x1` at 2:5",
    ), (
        'invalid_unicode_sequence_multiline',
        "var foo = 'foobar\\\nfail\\u12'",
        "Invalid unicode escape sequence `\\u12` at 2:5",
    ), (
        'long_invalid_string_truncated',
        "var foo = '1234567890abcdetruncated",
        'Unterminated string literal "\'1234567890abcde..." at 1:11',
    )
]

es5_comment_cases = [
    (
        'line_comment_whole',
        ('//comment\na = 5;\n',
         ['LINE_COMMENT //comment', 'ID a', 'EQ =', 'NUMBER 5', 'SEMI ;']),
    ), (
        'line_comment_trail',
        ('a//comment', ['ID a', 'LINE_COMMENT //comment']),
    ), (
        'block_comment_single',
        ('/***/b/=3//line',
         ['BLOCK_COMMENT /***/', 'ID b', 'DIVEQUAL /=',
          'NUMBER 3', 'LINE_COMMENT //line']),
    ), (
        'block_comment_multiline',
        ('/*\n * Copyright LGPL 2011 \n*/\na = 1;',
         ['BLOCK_COMMENT /*\n * Copyright LGPL 2011 \n*/',
          'ID a', 'EQ =', 'NUMBER 1', 'SEMI ;']),
    ), (
        # this will replace the standard test cases
        'section_7_comments',
        ("a = b\n/** **/\n\n/hi/s",
         ['ID a', 'EQ =', 'ID b', 'BLOCK_COMMENT /** **/', 'DIV /', 'ID hi',
          'DIV /', 'ID s'])
    )
]

# replace the section_7_comments test case
es5_all_cases = es5_cases[:-1] + es5_comment_cases

# double quote version
es5_error_cases_str_dq = [
    (n, arg.translate(swapquotes), msg.translate(swapquotes))
    for n, arg, msg in es5_error_cases_str
]

# single quote version
es5_error_cases_str_sq = [
    (n, arg, msg.translate({96: 39}))
    for n, arg, msg in es5_error_cases_str
]

es5_pos_cases = [
    (
        'single_line',
        """
        var foo = bar;  // line 1
        """, ([
            'var 1:0', 'foo 1:4', '= 1:8', 'bar 1:10', '; 1:13'
        ], [
            'var 1:1', 'foo 1:5', '= 1:9', 'bar 1:11', '; 1:14',
        ])
    ), (
        'multi_line',
        """
        var foo = bar;  // line 1


        var bar = baz;  // line 4
        """, ([
            'var 1:0', 'foo 1:4', '= 1:8', 'bar 1:10', '; 1:13',
            'var 4:28', 'bar 4:32', '= 4:36', 'baz 4:38', '; 4:41',
        ], [
            'var 1:1', 'foo 1:5', '= 1:9', 'bar 1:11', '; 1:14',
            'var 4:1', 'bar 4:5', '= 4:9', 'baz 4:11', '; 4:14',
        ])
    ), (
        'inline_comment',
        """
        // this is a comment  // line 1
        var foo = bar;  // line 2

        // another one  // line 4
        var bar = baz;  // line 5
        """, ([
            'var 2:32', 'foo 2:36', '= 2:40', 'bar 2:42', '; 2:45',
            'var 5:85', 'bar 5:89', '= 5:93', 'baz 5:95', '; 5:98',
        ], [
            'var 2:1', 'foo 2:5', '= 2:9', 'bar 2:11', '; 2:14',
            'var 5:1', 'bar 5:5', '= 5:9', 'baz 5:11', '; 5:14',
        ])
    ), (
        'block_comment',
        """
        /*
        This is a block comment
        */
        var foo = bar;  // line 4

        /* block single line */ // line 6
        var bar = baz;  // line 7

        /* oops */bar();  // line 9

          foo();
        """, ([
            'var 4:30', 'foo 4:34', '= 4:38', 'bar 4:40', '; 4:43',
            'var 7:91', 'bar 7:95', '= 7:99', 'baz 7:101', '; 7:104',
            'bar 9:128', '( 9:131', ') 9:132', '; 9:133',
            'foo 11:149', '( 11:152', ') 11:153', '; 11:154',
        ], [
            'var 4:1', 'foo 4:5', '= 4:9', 'bar 4:11', '; 4:14',
            'var 7:1', 'bar 7:5', '= 7:9', 'baz 7:11', '; 7:14',
            'bar 9:11', '( 9:14', ') 9:15', '; 9:16',
            'foo 11:3', '( 11:6', ') 11:7', '; 11:8',
        ])
    ), (
        'syntax_error_heading_comma',
        """
        var a;
        , b;
        """, ([
            'var 1:0', 'a 1:4', '; 1:5',
            ', 2:7', 'b 2:9', '; 2:10'
        ], [
            'var 1:1', 'a 1:5', '; 1:6',
            ', 2:1', 'b 2:3', '; 2:4'
        ])
    )
]


def run_lexer(value, lexer_cls):
    lexer = lexer_cls()
    lexer.input(value)
    return ['%s %s' % (token.type, token.value) for token in lexer]


def run_lexer_pos(value, lexer_cls):
    lexer = lexer_cls()
    lexer.input(textwrap.dedent(value).strip())
    tokens = list(lexer)
    return ([
        '%s %d:%d' % (token.value, token.lineno, token.lexpos)
        for token in tokens
    ], [
        '%s %d:%d' % (token.value, token.lineno, token.colno)
        for token in tokens
    ])



# -*- coding: utf-8 -*-
"""
Base class and function for making a walk through a given asttypes tree
possible.
"""

from __future__ import unicode_literals

from calmjs.parse.asttypes import Node
from calmjs.parse.ruletypes import Token
from calmjs.parse.ruletypes import Structure
from calmjs.parse.ruletypes import Layout
from calmjs.parse.ruletypes import LayoutChunk


def optimize_structure_handler(rule, handler):
    """
    Produce an "optimized" version of handler for the dispatcher to
    limit reference lookups.
    """

    def runner(walk, dispatcher, node):
        handler(dispatcher, node)
        return
        yield  # pragma: no cover

    return runner


def optimize_layout_handler(rule, handler):
    """
    Produce an "optimized" version of handler for the dispatcher to
    limit reference lookups.
    """

    def runner(walk, dispatcher, node):
        yield LayoutChunk(rule, handler, node)

    return runner


class Dispatcher(object):
    """
    Provide storage and lookup for the stored definitions and the
    handlers, which is documented by the constructor below.

    The instances of this class will be a callable that accept a single
    argument; when called with an object, the handling function, if any,
    will be returned to achieve the generation of output chunks.

    Calling on an instance of this object with a Rule (typically either
    a Token instance, or a Layout type/class object) will return the
    associated handler that was set up initially for a given instance.

    Accessing via a Node type as a key will return the definition that
    was initially set up for a given instance.

    The default implementation also provide a couple properties for ease
    of layout customization, which are the indent character and the
    newline character.  As certain users and/or platforms expect certain
    character sequences for these outputs, they can be specified in the
    constructor for this class, or be completely ignored by the specific
    layout handlers.

    An `error_handler` static method is provided by this class, which
    may be invoked (typically by the walk function) to signal an error.
    This default implementation will simply raise the exception passed.
    Subclasses may simply override this static method, or may simply
    assign a callable to an instance that accepts the same arguments
    for the handling of exceptions.

    While this class can be used (it was originally conceived) as a
    generic object that allow arbitrary assignments of arguments for
    consumption by layout functions, it's better to have a dedicated
    class that provide instance methods that plug into this.  See the
    modules inside ``calmjs.parse.handlers`` for various examples on
    how this could be set up.  To better maintain object purity, users
    of this class should not assign additional state attributes to
    instances of this class.
    """

    def __init__(
            self, definitions, token_handler,
            layout_handlers, deferrable_handlers,
            indent_str='  ', newline_str='\n'):
        """
        The constructor takes three arguments.

        definitions
            A mapping from the names of a Node to their definitions; a
            definition is a tuple of rules for describing how a
            particular Node should be rendered.  The Nodes are described
            in the asttypes module, while the Rules are described in the
            pptypes module.

        token_handler
            The handler that will deal with tokens.  It must be a
            callable that accepts four arguments

            token
                the token instance that will do the invocation
            dispatcher
                an instance of this class
            node
                a Node instance (from asttypes)
            value
                the value that was derived by the token based on its
                implementation.

        layout_handlers
            A map (dictionary) from Layout types to the handlers, which
            are callables that accepts these five arguments

            dispatcher
                an instance of this class
            node
                an asttypes.Node instance.
            before
                a value that was yielded by the previous token
            after
                a value to be yielded by the subsequent token
            prev
                the previously yielded layout token

        deferrable_handlers
            A map (dictionary) from Deferrable types to the handlers,
            which are callables that accepts these two arguments

            dispatcher
                an instance of this class
            node
                an asttypes.Node instance.

        indent_str
            The string used to indent a line with.  Default is '  '.
            This attribute will be provided as the property
            ``indent_str``.

        newline_str
            The string used for renderinga new line with.  Default is
            <LF> (line-feed, or '\\n').  This attribute will be provided
            as the property ``newline_str``.
        """

        self.__token_handler = token_handler
        self.__layout_handlers = {}
        self.__layout_handlers.update(layout_handlers)
        self.__deferrable_handlers = {}
        self.__deferrable_handlers.update(deferrable_handlers)
        self.__definitions = {}
        self.__definitions.update(definitions)
        self.__indent_str = indent_str
        self.__newline_str = newline_str

        self.__optimized_definitions = self.optimize()

    def optimize_definition(self, name, definition):
        rules = []
        for rule in definition:
            if isinstance(rule, type):
                if issubclass(rule, Structure):
                    handler = self.__layout_handlers.get(rule)
                    if handler:
                        rules.append(optimize_structure_handler(rule, handler))
                    continue
                elif issubclass(rule, Layout):
                    # a noop here so that the relevant chunk will be
                    # yielded for normalization by bulk-lookup.
                    handler = self.__layout_handlers.get(rule)
                    if handler:
                        rules.append(optimize_layout_handler(rule, handler))
                    continue
            if isinstance(rule, Token):
                value = (self.optimize_definition(
                    name, rule.value
                ) if isinstance(rule.value, tuple) else rule.value)
                rules.append(type(rule)(rule.attr, value, rule.pos))
                continue

            raise TypeError(
                "definition for '%s' contain unsupported rule (got: %r)" % (
                    name, rule))
        return rules

    def optimize(self):
        return {
            astname: self.optimize_definition(astname, definition)
            for astname, definition in self.__definitions.items()
        }

    def get_optimized_definition(self, node):
        """
        This is for getting at the definition for a particular asttype.
        """

        # The reason why the types were not used simply performance of
        # isinstance is bad, that the types are uniquely named, and that
        # they are always subclassed through the factory.  While working
        # at the type level is the correct method, the performance
        # penalties that it attracts however make this naive approach
        # more attractive.
        return self.__optimized_definitions[node.__class__.__name__]

    def __iter__(self):
        for item in self.__definitions.items():
            yield item

    def deferrable(self, rule):
        return self.__deferrable_handlers.get(type(rule), NotImplemented)

    def token(self, token, node, value, sourecepath_stack):
        if self.__token_handler:
            for fragment in self.__token_handler(
                    token, self, node, value, sourecepath_stack):
                yield fragment

    def layout(self, rule):
        """
        Get handler for this layout rule.
        """

        return self.__layout_handlers.get(rule, NotImplemented)

    @staticmethod
    def error_handler(exception, rule=None, node=None):
        raise exception

    @property
    def indent_str(self):
        return self.__indent_str

    @property
    def newline_str(self):
        return self.__newline_str

    @property
    def has_layout(self):
        return len(self.__layout_handlers) > 0


def walk(dispatcher, node, definition=None):
    """
    The default, standalone walk function following the standard
    argument ordering for the unparsing walkers.

    Arguments:

    dispatcher
        a Dispatcher instance, defined earlier in this module.  This
        instance will dispatch out the correct callable for the various
        object types encountered throughout this recursive function.

    node
        the starting Node from asttypes.

    definition
        a standalone definition tuple to start working on the node with;
        if none is provided, an initial definition will be looked up
        using the dispatcher with the node for the generation of output.

    While the dispatcher object is able to provide the lookup directly,
    this extra definition argument allow more flexibility in having
    Token subtypes being able to provide specific definitions also that
    may be required, such as the generation of optional rendering
    output.
    """

    # The inner walk function - this is actually exposed to the token
    # rule objects so they can also make use of it to process the node
    # with the dispatcher.

    nodes = []
    sourcepath_stack = [NotImplemented]
    has_layout = dispatcher.has_layout

    def _walk(dispatcher, node, definition=None, token=None):
        if not isinstance(node, Node):
            for fragment in dispatcher.token(
                    token, nodes[-1], node, sourcepath_stack):
                yield fragment
            return

        push = bool(node.sourcepath)
        if push:
            sourcepath_stack.append(node.sourcepath)
        nodes.append(node)

        if definition is None:
            definition = dispatcher.get_optimized_definition(node)

        for rule in definition:
            try:
                for chunk in rule(_walk, dispatcher, node):
                    yield chunk
            except Exception as e:
                yield dispatcher.error_handler(e, rule=rule, node=node)

        nodes.pop(-1)
        if push:
            sourcepath_stack.pop(-1)

    # Format layout markers are not handled immediately in the walk -
    # they will simply be buffered so that a collection of them can be
    # handled at once.
    def process_layouts(layout_rule_chunks, last_chunk, chunk):
        before_text = last_chunk.text if last_chunk else None
        after_text = chunk.text if chunk else None
        # the text that was yielded by the previous layout handler
        prev_text = None

        # While Layout rules in a typical definition are typically
        # interspersed with Tokens, certain assumptions with how the
        # Layouts are specified within there will fail when Tokens fail
        # to generate anything for any reason.  However, the dispatcher
        # instance will be able to accept and resolve a tuple of Layouts
        # to some handler function, so that a form of normalization can
        # be done.  For instance, an (Indent, Newline, Dedent) can
        # simply be resolved to no operations.  To achieve this, iterate
        # through the layout_rule_chunks and generate a normalized form
        # for the final handling to happen.

        # the preliminary stack that will be cleared whenever a
        # normalized layout rule chunk is generated.
        lrcs_stack = []

        # first pass: generate both the normalized/finalized lrcs.
        for lrc in layout_rule_chunks:
            lrcs_stack.append(lrc)

            # check every single chunk from left to right...
            for idx in range(len(lrcs_stack)):
                rule = tuple(lrc.rule for lrc in lrcs_stack[idx:])
                handler = dispatcher.layout(rule)
                if handler is not NotImplemented:
                    # not manipulating lrsc_stack from within the same
                    # for loop that it is being iterated upon
                    break
            else:
                # which continues back to the top of the outer for loop
                continue

            # So a handler is found from inside the rules; extend the
            # chunks from the stack that didn't get normalized, and
            # generate a new layout rule chunk.
            lrcs_stack[:] = lrcs_stack[:idx]
            lrcs_stack.append(LayoutChunk(
                rule, handler,
                layout_rule_chunks[idx].node,
            ))

        # second pass: now the processing can be done.
        for lr_chunk in lrcs_stack:
            gen = lr_chunk.handler(
                dispatcher, lr_chunk.node, before_text, after_text, prev_text)
            if not gen:
                continue
            for chunk_from_layout in gen:
                yield chunk_from_layout
                prev_text = chunk_from_layout.text

    # The top level walker implementation
    def walk():
        last_chunk = None
        layout_rule_chunks = []

        for chunk in _walk(dispatcher, node, definition):
            if isinstance(chunk, LayoutChunk):
                layout_rule_chunks.append(chunk)
            else:
                if has_layout:
                    # process layout rule chunks that had been cached.
                    for chunk_from_layout in process_layouts(
                            layout_rule_chunks, last_chunk, chunk):
                        yield chunk_from_layout
                layout_rule_chunks[:] = []
                yield chunk
                last_chunk = chunk

        if has_layout:
            # process the remaining layout rule chunks.
            for chunk_from_layout in process_layouts(
                    layout_rule_chunks, last_chunk, None):
                yield chunk_from_layout

    for chunk in walk():
        yield chunk



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import unittest
import base64
import json
from io import StringIO
from os.path import join
from tempfile import mktemp

from calmjs.parse.exceptions import ECMASyntaxError
from calmjs.parse.asttypes import Node
from calmjs.parse.ruletypes import Attr
from calmjs.parse.ruletypes import Text
from calmjs.parse.unparsers.base import BaseUnparser
from calmjs.parse import io


class IOTestCase(unittest.TestCase):

    def test_read(self):
        # given that there is no generic parser, emulate one like so
        def parser(text):
            result = Node()
            result.raw = text
            return result

        stream = StringIO('var foo = "bar";')
        node = io.read(parser, stream)
        self.assertIsNone(node.sourcepath)

        stream.name = 'somefile.js'
        node = io.read(parser, stream)
        self.assertEqual(node.sourcepath, 'somefile.js')

    def test_read_callable(self):
        # given that there is no generic parser, emulate one like so
        def parser(text):
            result = Node()
            result.raw = text
            return result

        stream = StringIO('var foo = "bar";')

        def close():
            stream._closed = True

        def wrapper():
            stream.close = close
            return stream

        stream = StringIO('var foo = "bar";')
        io.read(parser, wrapper)
        self.assertTrue(stream._closed)

    def test_read_error(self):
        # given that there is no generic parser, emulate one like so
        def parser(text):
            raise ECMASyntaxError('Illegal input')

        def not_a_parser(text):
            raise Exception('Illegal input')

        stream = StringIO('var foo = "bar";')
        stream.name = 'somefile.js'
        with self.assertRaises(Exception) as e:
            io.read(parser, stream)

        self.assertEqual("Illegal input in 'somefile.js'", e.exception.args[0])

        stream = StringIO('var foo = "bar";')
        with self.assertRaises(Exception) as e:
            io.read(not_a_parser, stream)

        self.assertNotEqual(
            "Illegal input in 'somefile.js'", e.exception.args[0])

    def test_write_no_sourcemap(self):
        root = mktemp()
        definitions = {'Node': (
            Attr(attr='left'), Attr(attr='op'), Attr(attr='right'),)}

        # the program node; attributes are assigned to mimic a real one
        program = Node()
        program.left, program.op, program.right = ('foo', '=', 'true')
        program.sourcepath = join(root, 'original.js')
        program._token_map = {
            'foo': [(0, 1, 1)],
            '=': [(4, 1, 5)],
            'true': [(6, 1, 7)],
        }

        output_stream = StringIO()
        output_stream.name = join(root, 'processed.js')

        unparser = BaseUnparser(definitions)
        io.write(unparser, program, output_stream)
        self.assertEqual('foo=true', output_stream.getvalue())

    def test_write_sourcemap(self):
        root = mktemp()
        definitions = {'Node': (
            Attr(attr='left'), Attr(attr='op'), Attr(attr='right'),)}

        # the program node; attributes are assigned to mimic a real one
        program = Node()
        program.left, program.op, program.right = ('foo', '=', 'true')
        program.sourcepath = join(root, 'original.js')
        program._token_map = {
            'foo': [(0, 1, 1)],
            '=': [(4, 1, 5)],
            'true': [(6, 1, 7)],
        }

        # streams
        output_stream = StringIO()
        output_stream.name = join(root, 'processed.js')
        sourcemap_stream = StringIO()
        sourcemap_stream.name = join(root, 'processed.js.map')

        unparser = BaseUnparser(definitions)
        io.write(unparser, program, output_stream, sourcemap_stream)

        sourcemap = json.loads(sourcemap_stream.getvalue())
        self.assertEqual({
            "version": 3,
            "sources": ["original.js"],
            "names": [],
            "mappings": "AAAA,GAAI,CAAE",
            "file": "processed.js"
        }, sourcemap)
        self.assertEqual(
            'foo=true\n//# sourceMappingURL=processed.js.map\n',
            output_stream.getvalue())

    def test_write_sourcemap_omitted(self):
        root = mktemp()
        definitions = {'Node': (
            Attr(attr='left'), Attr(attr='op'), Attr(attr='right'),)}

        # the program node; attributes are assigned to mimic a real one
        program = Node()
        program.left, program.op, program.right = ('foo', '=', 'true')
        program.sourcepath = join(root, 'original.js')
        program._token_map = {
            'foo': [(0, 1, 1)],
            '=': [(4, 1, 5)],
            'true': [(6, 1, 7)],
        }

        # streams
        output_stream = StringIO()
        output_stream.name = join(root, 'processed.js')
        sourcemap_stream = StringIO()
        sourcemap_stream.name = join(root, 'processed.js.map')

        unparser = BaseUnparser(definitions)
        io.write(
            unparser, program, output_stream, sourcemap_stream,
            source_mapping_url=None)

        sourcemap = json.loads(sourcemap_stream.getvalue())
        self.assertEqual({
            "version": 3,
            "sources": ["original.js"],
            "names": [],
            "mappings": "AAAA,GAAI,CAAE",
            "file": "processed.js"
        }, sourcemap)
        self.assertEqual(
            'foo=true',
            output_stream.getvalue())

    def test_write_multiple(self):
        root = mktemp()
        definitions = {'Node': (
            Attr(attr='left'), Text(value=' '),
            Attr(attr='op'), Text(value=' '),
            Attr(attr='right'), Text(value=';'),
        )}

        # the program node; attributes are assigned to mimic a real one
        program1 = Node()
        program1.left, program1.op, program1.right = ('foo', '=', 'true')
        program1.sourcepath = join(root, 'program1.js')
        program1._token_map = {
            'foo': [(0, 1, 1)],
            '=': [(4, 1, 5)],
            'true': [(6, 1, 7)],
        }
        program2 = Node()
        program2.left, program2.op, program2.right = ('bar', '=', 'false')
        program2.sourcepath = join(root, 'program2.js')
        program2._token_map = {
            'bar': [(0, 1, 1)],
            '=': [(4, 1, 5)],
            'false': [(6, 1, 7)],
        }

        # streams
        output_stream = StringIO()
        output_stream.name = join(root, 'packed.js')
        sourcemap_stream = StringIO()
        sourcemap_stream.name = join(root, 'packed.js.map')

        unparser = BaseUnparser(definitions)
        io.write(
            unparser, [program1, program2], output_stream, sourcemap_stream,
            source_mapping_url=None)

        self.assertEqual('foo = true;bar = false;', output_stream.getvalue())

        sourcemap = json.loads(sourcemap_stream.getvalue())
        self.assertEqual({
            "version": 3,
            "sources": ["program1.js", "program2.js"],
            "names": [],
            "mappings": "AAAA,WCAA",
            "file": "packed.js"
        }, sourcemap)

    def test_write_callables(self):
        closed = []

        class Stream(StringIO):
            # don't actually close the stream so it can be read later by
            # the tests
            def close(self):
                closed.append(self)

        # streams
        root = mktemp()
        output_stream = Stream()
        output_stream.name = join(root, 'packed.js')
        sourcemap_stream = Stream()
        sourcemap_stream.name = join(root, 'packed.js.map')

        def f_output_stream():
            return output_stream

        def f_sourcemap_stream():
            return sourcemap_stream

        definitions = {'Node': (
            Attr(attr='left'), Text(value=' '),
            Attr(attr='op'), Text(value=' '),
            Attr(attr='right'), Text(value=';'),
        )}

        # the program node; attributes are assigned to mimic a real one
        program1 = Node()
        program1.left, program1.op, program1.right = ('foo', '=', 'true')
        program1.sourcepath = join(root, 'program1.js')
        program1._token_map = {
            'foo': [(0, 1, 1)],
            '=': [(4, 1, 5)],
            'true': [(6, 1, 7)],
        }
        program2 = Node()
        program2.left, program2.op, program2.right = ('bar', '=', 'false')
        program2.sourcepath = join(root, 'program2.js')
        program2._token_map = {
            'bar': [(0, 1, 1)],
            '=': [(4, 1, 5)],
            'false': [(6, 1, 7)],
        }

        unparser = BaseUnparser(definitions)
        io.write(
            unparser, [program1, program2],
            f_output_stream, f_sourcemap_stream,
            source_mapping_url=None)

        self.assertIn(output_stream, closed)
        self.assertIn(sourcemap_stream, closed)
        self.assertEqual('foo = true;bar = false;', output_stream.getvalue())

        sourcemap = json.loads(sourcemap_stream.getvalue())
        self.assertEqual({
            "version": 3,
            "sources": ["program1.js", "program2.js"],
            "names": [],
            "mappings": "AAAA,WCAA",
            "file": "packed.js"
        }, sourcemap)

    def test_write_same_stream_callable(self):
        # streams
        root = mktemp()
        output_stream = StringIO()
        output_stream.name = join(root, 'packed.js')
        called = []
        closed = []

        def close():
            closed.append(True)

        output_stream.close = close

        def f_output_stream():
            called.append(True)
            return output_stream

        definitions = {'Node': (
            Attr(attr='text'), Text(value=';'),
        )}

        # the program node; attributes are assigned to mimic a real one
        program = Node()
        program.text = 'hello'
        program.sourcepath = join(root, 'program.js')
        program._token_map = {'hello': [(0, 1, 1)]}

        unparser = BaseUnparser(definitions)
        io.write(unparser, [program], f_output_stream, f_output_stream)

        self.assertEqual(1, len(called))
        self.assertEqual(1, len(closed))
        output = output_stream.getvalue()
        self.assertIn('hello', output)
        self.assertNotIn('program.js', output)
        # since output stream is a StringIO, default to utf8 encoding
        self.assertIn('data:application/json;base64;charset=utf8', output)
        # decode the base64 string
        self.assertEqual({
            "version": 3,
            "sources": ["program.js"],
            "names": [],
            "mappings": "AAAA",
            "file": "packed.js"
        }, json.loads(base64.b64decode(
            output.splitlines()[-1].split(',')[-1].encode('utf8')
        ).decode('utf8')))

    def test_write_error_handled_callable_closed(self):
        # streams
        root = mktemp()
        output_stream = StringIO()
        output_stream.name = join(root, 'packed.js')
        closed = []

        def close():
            closed.append(True)

        output_stream.close = close

        def f_output_stream():
            return output_stream

        def f_error():
            raise IOError('some error happened')

        definitions = {'Node': (
            Attr(attr='text'), Text(value=';'),
        )}

        # the program node; attributes are assigned to mimic a real one
        program = Node()
        program.text = 'hello'
        program.sourcepath = join(root, 'program.js')
        program._token_map = {'hello': [(0, 1, 1)]}

        unparser = BaseUnparser(definitions)
        with self.assertRaises(IOError):
            io.write(unparser, [program], f_output_stream, f_error)

        self.assertEqual(1, len(closed))
        self.assertEqual('hello;', output_stream.getvalue())
        self.assertNotIn('program.js', output_stream.getvalue())

    def test_write_wrong_type(self):
        stream = StringIO()
        unparser = BaseUnparser({})
        with self.assertRaises(TypeError):
            io.write(unparser, [], stream)

        with self.assertRaises(TypeError):
            io.write(unparser, '', stream)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import unittest

from calmjs.parse.asttypes import Node
from calmjs.parse.unparsers.base import logger
from calmjs.parse.unparsers.base import BaseUnparser
from calmjs.parse.unparsers.walker import Dispatcher
from calmjs.parse.handlers.core import token_handler_str_default

from calmjs.parse.testing.util import setup_logger


class BaseUnparserTestCase(unittest.TestCase):

    def test_prewalk_fail(self):
        definitions = {}
        unparser = BaseUnparser(definitions)
        # can't lookup an empty definition
        with self.assertRaises(KeyError):
            self.assertEqual([], list(unparser(Node())))

    def test_minimum_definition(self):
        definitions = {'Node': ()}
        unparser = BaseUnparser(definitions)
        self.assertEqual([], list(unparser(Node())))

    def test_prewalk_hooking(self):
        results = {}

        def prewalk_dummy(dispatcher, node):
            results.update({'dispatcher': dispatcher, 'node': node})
            return node

        definitions = {'Node': ()}
        unparser = BaseUnparser(definitions, prewalk_hooks=[prewalk_dummy])
        self.assertEqual(results, {})
        # invoke complete run to trigger prewalk hook.
        root = Node()
        self.assertEqual([], list(unparser(root)))
        self.assertTrue(isinstance(results['dispatcher'], Dispatcher))
        self.assertTrue(results['node'], root)

    def test_called_prewalk_via_rules(self):
        results = {}

        def prewalk_dummy(dispatcher, node):
            results.update({'dispatcher': dispatcher, 'node': node})
            return node

        def rule():
            return {'prewalk_hooks': (prewalk_dummy,)}

        definitions = {'Node': ()}
        unparser = BaseUnparser(definitions, rules=(rule,))
        # invoke complete run to trigger prewalk hook.
        root = Node()
        self.assertEqual([], list(unparser(root)))
        self.assertTrue(isinstance(results['dispatcher'], Dispatcher))
        self.assertTrue(results['node'], root)

    def test_called_prewalk_multicall(self):
        prewalk = []

        def rule():
            prewalk.append(True)
            return {}

        root = Node()
        definitions = {'Node': ()}
        unparser = BaseUnparser(definitions, rules=(rule,))
        # invoke complete run to trigger prewalk hook.
        self.assertEqual(len(prewalk), 0)
        self.assertEqual([], list(unparser(root)))
        self.assertEqual(len(prewalk), 1)
        self.assertEqual([], list(unparser(root)))
        self.assertEqual(len(prewalk), 2)

    def test_token_handler_default(self):
        stream = setup_logger(self, logger)
        definitions = {}
        unparser = BaseUnparser(definitions)
        token_handler, layout_handlers, deferrable_handlers, prewalk_hooks = (
            unparser.setup())
        self.assertIs(token_handler, token_handler_str_default)
        self.assertIn(
            "DEBUG 'BaseUnparser' instance has no token_handler specified; "
            "default handler 'token_handler_str_default' activate",
            stream.getvalue())

    def test_token_handler_setup_manual(self):
        stream = setup_logger(self, logger)
        definitions = {}
        unparser = BaseUnparser(definitions, token_handler_str_default)
        token_handler, layout_handlers, deferrable_handlers, prewalk_hooks = (
            unparser.setup())
        self.assertIs(token_handler, token_handler_str_default)
        self.assertIn(
            "DEBUG 'BaseUnparser' instance using manually specified "
            "token_handler 'token_handler_str_default'", stream.getvalue()
        )

    def test_token_handler_setup_with_rules(self):
        def rule1():
            def handler1():
                "handler1"
            return {'token_handler': handler1}

        def rule2():
            def handler2():
                "handler2"
            return {'token_handler': handler2}

        def custom_handler():
            "custom handler"

        stream = setup_logger(self, logger)
        definitions = {}
        unparser = BaseUnparser(definitions, rules=(rule1, rule2))
        token_handler, layout_handlers, deferrable_handlers, prewalk_hooks = (
            unparser.setup())
        self.assertEqual(token_handler.__name__, 'handler2')

        self.assertIn(
            "DEBUG rule 'rule1' specified a token_handler 'handler1'",
            stream.getvalue()
        )
        self.assertIn(
            "WARNING rule 'rule2' specified a new token_handler 'handler2', "
            "overriding previously assigned token_handler 'handler1'",
            stream.getvalue()
        )

        unparser = BaseUnparser(
            definitions, token_handler=custom_handler, rules=(rule1, rule2))
        token_handler, layout_handlers, deferrable_handlers, prewalk_hooks = (
            unparser.setup())
        self.assertIs(token_handler, custom_handler)
        self.assertIn(
            "INFO manually specified token_handler 'custom_handler' to the "
            "'BaseUnparser' instance will override rule derived token_handler "
            "'handler2'",
            stream.getvalue()
        )



# -*- coding: utf-8 -*-
"""
Generic io functions for use with parsers.
"""

from itertools import chain
try:
    from collections.abc import Iterable
except ImportError:  # pragma: no cover
    from collections import Iterable
from calmjs.parse.asttypes import Node
from calmjs.parse import sourcemap
from calmjs.parse.exceptions import ECMASyntaxError
from calmjs.parse.utils import repr_compat


def read(parser, stream):
    """
    Return an AST from the input ES5 stream.

    Arguments

    parser
        A parser instance.
    stream
        Either a stream object or a callable that produces one.  The
        stream object to read from; its 'read' method will be invoked.

        If a callable was provided, the 'close' method on its return
        value will be called to close the stream.
    """

    source = stream() if callable(stream) else stream
    try:
        text = source.read()
        stream_name = getattr(source, 'name', None)
        try:
            result = parser(text)
        except ECMASyntaxError as e:
            error_name = repr_compat(stream_name or source)
            raise type(e)('%s in %s' % (str(e), error_name))
    finally:
        if callable(stream):
            source.close()

    result.sourcepath = stream_name
    return result


def write(
        unparser, nodes, output_stream, sourcemap_stream=None,
        sourcemap_normalize_mappings=True,
        sourcemap_normalize_paths=True,
        source_mapping_url=NotImplemented):
    """
    Write out the node using the unparser into an output stream, and
    optionally the sourcemap using the sourcemap stream.

    Ideally, file objects should be passed to the *_stream arguments, so
    that the name resolution built into the sourcemap builder function
    will be used.  Also, if these file objects are opened using absolute
    path arguments, enabling the sourcemap_normalize_paths flag will
    have all paths normalized to their relative form.

    If the provided streams are not anchored on the filesystem, or that
    the provide node was generated from a string or in-memory stream,
    the generation of the sourcemap should be done using the lower level
    `write` function provided by the sourcemap module, which this method
    wraps.  Alternatively, the top level node should have its sourcepath
    set to path that this node originated from.

    Arguments

    unparser
        An unparser instance.
    nodes
        The Node or list of Nodes to stream to the output stream with
        the unparser.
    output_stream
        Either a stream object or a callable that produces one.  The
        stream object to write to; its 'write' method will be invoked.

        If a callable was provided, the 'close' method on its return
        value will be called to close the stream.
    sourcemap_stream
        If one is provided, the sourcemap will be written out to it.
        Like output_stream, it could also be a callable and be handled
        in the same manner.

        If this argument is the same as output_stream (note: the return
        value of any callables are not compared), the stream object that
        is the same as the output_stream will be used for writing out
        the source map, and the source map will instead be encoded as a
        'data:application/json;base64,' URL.
    sourcemap_normalize_mappings
        Flag for the normalization of the sourcemap mappings; Defaults
        to True to enable a reduction in output size.
    sourcemap_normalize_paths
        If set to true, all absolute paths will be converted to the
        relative form when the sourcemap is generated, if all paths
        provided are in the absolute form.

        Defaults to True to enable a reduction in output size.
    source_mapping_url
        If unspecified, the default derived path will be written as a
        sourceMappingURL comment into the output stream.  If explicitly
        specified with a value, that will be written instead.  Set to
        None to disable this.
    """

    closer = []

    def get_stream(stream):
        if callable(stream):
            result = stream()
            closer.append(result.close)
        else:
            result = stream
        return result

    def cleanup():
        for close in reversed(closer):
            close()

    chunks = None
    if isinstance(nodes, Node):
        chunks = unparser(nodes)
    elif isinstance(nodes, Iterable):
        raw = [unparser(node) for node in nodes if isinstance(node, Node)]
        if raw:
            chunks = chain(*raw)

    if not chunks:
        raise TypeError('must either provide a Node or list containing Nodes')

    try:
        out_s = get_stream(output_stream)
        sourcemap_stream = (
            out_s if sourcemap_stream is output_stream else sourcemap_stream)
        mappings, sources, names = sourcemap.write(
            chunks, out_s, normalize=sourcemap_normalize_mappings)
        if sourcemap_stream:
            sourcemap_stream = get_stream(sourcemap_stream)
            sourcemap.write_sourcemap(
                mappings, sources, names, out_s, sourcemap_stream,
                normalize_paths=sourcemap_normalize_paths,
                source_mapping_url=source_mapping_url,
            )
    finally:
        cleanup()



# -*- coding: utf-8 -*-
"""
Description for the extractor unparser
"""

from __future__ import unicode_literals

import operator
from ast import literal_eval
from collections import namedtuple
from collections import defaultdict
from math import floor
from logging import getLogger
try:
    from collections.abc import MutableSequence
except ImportError:  # pragma: no cover
    from collections import MutableSequence

from calmjs.parse.asttypes import (
    Array,
    Assign,
    BinOp,
    Boolean,
    FunctionCall,
    Number,
    Null,
    Object,
    String,
    UnaryExpr,
    nodetype,
)
from calmjs.parse.ruletypes import (
    # Space,
    PushScope,
    PopScope,
)
from calmjs.parse.ruletypes import (
    # is_empty,

    Attr,
    Token,
    Iter,
    Text,
    Optional,
    JoinAttr,
    Operator,
)
from calmjs.parse.ruletypes import (
    Literal,
    Declare,
    Resolve,
    ResolveFuncName,
)
from calmjs.parse.unparsers.base import BaseUnparser
from calmjs.parse.unparsers import walker
from calmjs.parse.utils import str


inf = float('inf')
logger = getLogger(__name__)

# The primary fragment type; token_handler_extractor yields these
ExtractedFragment = namedtuple('ExtractedFragment', [
    'value',
    'node',
    'folded_type',
])
# The secondary fragment type; may be used as value so that the
# token_handler_extractor be able to produce a folded_type
FoldedFragment = namedtuple('FoldedFragment', [
    'value',
    'folded_type',
])


# See ECMA-262 5.1 Edition, Section 9
# Note that this parser does not provide an undefined token or global
# binding, it's currently not handled.

def value_to_str(value):
    """
    Since the design eagerly convert and discard because it was felt
    that tracking everything all at once being too cumbersome, there
    will be situations where this conversion from Python values back
    to JavaScript equivalent string expression...

    This is currently used for converting converted values enclosed in
    Array back to the JavaScript string converted types; not to be used
    in subsequent methods due to subtle differences with the conversion
    of null.
    """

    if value is None:  # null
        return ''
    elif value is True:  # true
        return 'true'
    elif value is False:  # false
        return 'false'
    elif isinstance(value, (int, float)):
        return str(value)
    elif isinstance(value, str):
        return value
    elif isinstance(value, list):
        return ','.join([value_to_str(v) for v in value])
    elif isinstance(value, dict):
        return '[object Object]'
    raise TypeError(
        '%r cannot be converted to a value in the base ECMAScript '
        'language type' % value)


def to_primitive(fragment, hint):
    # The hint argument is the PreferredType
    if (issubclass(fragment.folded_type, Array) or
            issubclass(fragment.folded_type, Object)):
        value = value_to_str(fragment.value)
        if issubclass(hint, Number):
            return FoldedFragment(
                to_number(FoldedFragment(value, String)), Number)
        # everything else?
        return FoldedFragment(value, String)
    # no conversion for other types.
    return fragment


def to_boolean(fragment):
    if issubclass(fragment.folded_type, Array):
        return True
    elif issubclass(fragment.folded_type, Boolean):
        return fragment.value
    elif issubclass(fragment.folded_type, Number):
        return fragment.value not in (0, 'NaN')
    elif issubclass(fragment.folded_type, String):
        return bool(fragment.value)
    elif issubclass(fragment.folded_type, Object):
        return True
    return False


def to_number(fragment):
    def from_str(value):
        # probably not exactly spec conformant, but close enough.
        maybe_hex = value.strip()
        if maybe_hex.startswith('0x'):
            cleaned = maybe_hex
        else:
            cleaned = maybe_hex.lstrip('0')
        if cleaned == '':
            return 0
        try:
            number = literal_eval(cleaned)
        except (ValueError, SyntaxError):
            return 'NaN'
        if not isinstance(number, (int, float)):
            return 'NaN'
        return number

    if issubclass(fragment.folded_type, Array):
        return to_primitive(fragment, Number).value
    elif issubclass(fragment.folded_type, Boolean):
        return int(fragment.value)
    elif issubclass(fragment.folded_type, Number):
        return fragment.value
    elif issubclass(fragment.folded_type, Null):
        return 0
    elif issubclass(fragment.folded_type, String):
        return from_str(fragment.value)
    elif issubclass(fragment.folded_type, Object):
        return to_primitive(fragment, Number).value
    return 'NaN'


def to_integer(fragment):
    value = to_number(fragment)
    if value == 'NaN':
        return 0
    if value in (0, inf):
        return value
    return (1 if abs(value) == value else -1) * int(floor(abs(value)))


def to_uint32(fragment):
    value = to_integer(fragment)
    if value in ('NaN', 0, inf):
        return value
    posint = (1 if abs(value) == value else -1) * int(floor(abs(value)))
    return posint % 2 ** 32


def to_int32(fragment):
    int32bit = to_uint32(fragment)
    if int32bit >= 2 ** 31:
        return int32bit - 2 ** 32
    else:
        return int32bit


def to_string(fragment):
    # can't quite use value_to_str because that's for values enclosed in
    # another object.
    if issubclass(fragment.folded_type, Array):
        return to_string(to_primitive(fragment, String))
    elif issubclass(fragment.folded_type, Boolean):
        return 'true' if fragment.value else 'false'
    elif issubclass(fragment.folded_type, Number):
        return str(fragment.value)
    elif issubclass(fragment.folded_type, Null):
        return 'null'
    elif issubclass(fragment.folded_type, String):
        return fragment.value
    elif issubclass(fragment.folded_type, Object):
        return to_string(to_primitive(fragment, String))

    # For other unsupported values; if the fragment value is a string
    # type, assume it is some kind of resolvable identifier that may be
    # used as a format placeholder token.
    elif isinstance(fragment.value, str):
        return '{%s}' % fragment.value
    # Function Calls have been destructured, but given that it is
    # impractical to reapply as is, just replace with an anonymous
    # field.
    elif issubclass(fragment.folded_type, FunctionCall):
        return '{}'
    # Otherwise, return an empty string.
    return ''


class Assignment(tuple):
    """
    Denote a thing to be assigned.
    """

    def __new__(_cls, key, value):
        return tuple.__new__(_cls, (key, value))

    @property
    def key(self):
        return (
            self[0].value
            if isinstance(self[0], ExtractedFragment) else
            self[0]
        )

    @property
    def value(self):
        return (
            self[1].value
            if isinstance(self[1], ExtractedFragment) else
            self[1]
        )

    def __iter__(self):
        yield self.key
        yield self.value

    def __repr__(self):
        return '(%r, %r)' % (self.key, self.value)


class AssignmentList(MutableSequence):
    """
    A mapping for a list of assignments.
    """

    def __init__(self, *a):
        self.__seq = []
        if len(a) == 1 and isinstance(a[0], AssignmentList):
            self.__seq.extend(a[0])
        else:
            self.extend(a)

    def normalize(self, value):
        if isinstance(value, Assignment):
            return value
        elif isinstance(value, list):
            if len(value) == 2:
                return Assignment(value[0], value[1])
        raise ValueError('%r cannot be converted to an Assignment' % (value,))

    def __getitem__(self, index):
        return self.__seq[index]

    def __setitem__(self, index, value):
        self.__seq[index] = self.normalize(value)

    def __delitem__(self, index):
        self.__seq.__delitem__(index)

    def insert(self, index, value):
        self.__seq.insert(index, self.normalize(value))

    # def __iter__(self):
    #     return iter(self.__seq)

    def __len__(self):
        return len(self.__seq)

    def __repr__(self):
        return repr(self.__seq)

    def __eq__(self, other):
        return self.__seq == other


# Defining custom ruletype tokens in this module instead of the main
# ruletypes module, simply due to the fact that what is being yielded is
# currently not compatible at all with the default token handler that
# yields stream fragments or layout chunks.
#
# Formalizng this API will be useful, it will be quite a lot of work and
# rather difficult without very strict pre-compile static typing - i.e.
# not possible in Python.

class GroupAs(Token):
    """
    A special token where the provided attr must be a tuple of tokens
    or rules, such that those rules will result in a discrete container
    of all the tokens being yielded, instead of being flattened.

    Provides a common build_items method which may be used to produce
    the list of items.
    """

    def build_items(self, walk, dispatcher, node):
        return (
            item
            # TODO the name argument should really be the name of the
            # type of this node... to help with debugging.
            for attr in dispatcher.optimize_definition('', self.attr)
            for item in attr(walk, dispatcher, node)
        )


class GroupAsAssignment(GroupAs):
    """
    Simply group what was produced by the token as a single assignment.

    There must be two distinct elements, otherwise it will fail.
    """

    def __call__(self, walk, dispatcher, node):
        yield next(dispatcher.token(None, node, AssignmentList(
            list(self.build_items(walk, dispatcher, node))), None))


class GroupAsList(GroupAs):
    """
    Take the processed chunks and remove the context information by
    taking the stored values and return a list.
    """

    def __call__(self, walk, dispatcher, node):
        yield next(dispatcher.token(None, node, [
            v.value for v in self.build_items(walk, dispatcher, node)], None))


class GroupAsMap(GroupAs):
    """
    Take the processed chunks and remove the context information by
    taking the stored values and return a mapping.  Ensure that the
    AssignmentList types are processed for the Assignment entries, and
    extract the unmatched entries into its own list under the
    NotImplemented key.
    """

    def __call__(self, walk, dispatcher, node):
        result = {}
        misc = defaultdict(list)
        for item in self.build_items(walk, dispatcher, node):
            if isinstance(item.value, AssignmentList):
                # in the case of certain in place statement that yielded
                # multiple assignments within the same scope
                result.update(item.value)
            else:
                # if what was produced was not already a list of
                # assignments, chuck it in with the misc list.
                misc[nodetype(item.node)].append(item.value)

        result.update(misc)
        yield next(dispatcher.token(None, node, result, None))


class GroupAsStr(GroupAs):
    """
    Take the processed chunks and extract the value, conver them to text
    and join them together.
    """

    def __call__(self, walk, dispatcher, node):
        joiner = self.value if self.value else ''
        yield next(dispatcher.token(None, node, joiner.join(
            str(s.value) for s in self.build_items(walk, dispatcher, node)
        ), None))


class AttrListAssignment(Attr):
    """
    Specialized attr to produce a ListAssignment.

    The attr argument for the constructor in this case should be a
    2-tuple, first element being the attribute name of the LHS, second
    element being the same for RHS.  Hence LHS becomes the "key", and
    RHS becomes the "value".

    During the production of the yielded Assignmentlist, if the RHS is
    an assignment, the value of the rightmost RHS value will be yielded,
    with the key produced from LHS, along with any other assignment that
    was produced from RHS.
    """

    def __call__(self, walk, dispatcher, node):
        lhs = getattr(node, self.attr[0])
        rhs = getattr(node, self.attr[1])

        def standard_walk():
            # discard all chunk information here?
            # TODO probably should track them to aid with debugging?
            for chunk in walk(dispatcher, lhs, token=self):
                yield chunk.value
            for chunk in walk(dispatcher, rhs, token=self):
                yield chunk.value

        chunks = list(standard_walk())

        if isinstance(rhs, Assign):
            # if RHS is an assign node, yield LHS with the final chunk
            # value produced with the RHS through the walk, then the
            # rest of the chunks produced by the walk.
            yield next(dispatcher.token(None, node, AssignmentList(
                Assignment(chunks[0], chunks[1][-1][1]), *chunks[1]), None))
        else:
            yield next(dispatcher.token(None, node, AssignmentList(
                Assignment(*chunks)), None))


class OpDisambiguate(Token):
    """
    A special token that deals with nodes with an `op` attribute, such
    that it will be used to determine the handling to use through the
    mapping specified with the value provided for the construction of
    this token.

    A default token should be specified with the key `NotImplemented`.
    """

    def __call__(self, walk, dispatcher, node):
        definition = dispatcher.optimize_definition('', self.value.get(
            node.op, self.value.get(NotImplemented, ())))
        for rule in definition:
            for chunk in rule(walk, dispatcher, node):
                yield chunk


class GroupAsBinOp(GroupAs):
    """
    The generic binop
    """

    def _next_one(self, walk, dispatcher, node):
        gen = walk(dispatcher, node, definition=None)
        result = next(gen)
        try:
            fail = next(gen)
        except StopIteration:
            return result
        else:
            # TODO exception type
            raise ValueError(
                "Ruletype token %r unable to process output produced by the "
                "definition used on <%s @%s:%s ...>, as it yielded more than "
                "one fragment (first two fragments are %r and %r)" % (
                    type(self).__name__,
                    # manually construct a zero depth with line number
                    type(node).__name__,
                    '?' if node.lineno is None else node.lineno,
                    '?' if node.colno is None else node.colno,
                    result,
                    fail,
                )
            )

    def __call__(self, walk, dispatcher, node):
        if not isinstance(node, BinOp):
            raise TypeError(
                "Ruletype token %r expects a 'BinOp', got %r" % (
                    type(self).__name__, type(node).__name__))

        lhs = self._next_one(walk, dispatcher, node.left)
        rhs = self._next_one(walk, dispatcher, node.right)
        yield next(dispatcher.token(None, node, self.binop(lhs, rhs), None))

    def binop(self, lhs, rhs):
        """
        Default implementation is the most common form.
        """

        # assumes to be ExtractedFragments
        lhs_value = to_number(lhs)
        rhs_value = to_number(rhs)
        if lhs_value == 'NaN' or rhs_value == 'NaN':
            return FoldedFragment('NaN', Number)
        else:
            try:
                return FoldedFragment(self.op(lhs_value, rhs_value), Number)
            except ArithmeticError:
                return FoldedFragment('NaN', Number)

    def op(self, lhs, rhs):
        raise NotImplementedError


class GroupAsBinOpPlus(GroupAsBinOp):
    """
    For BinOp with op = '+'
    """

    def binop(self, lhs, rhs):
        """
        The plus operator is a bit different from the rest.
        """

        if issubclass(lhs.folded_type, Number) and issubclass(
                rhs.folded_type, Number):
            return FoldedFragment(lhs.value + rhs.value, Number)
        else:
            return FoldedFragment(to_string(lhs) + to_string(rhs), String)


class GroupAsBinOpMinus(GroupAsBinOp):
    """
    For BinOp with op = '-'
    """

    op = operator.sub


class GroupAsBinOpMult(GroupAsBinOp):
    """
    For BinOp with op = '*'
    """

    op = operator.mul


class GroupAsBinOpDiv(GroupAsBinOp):
    """
    For BinOp with op = '*'
    """

    op = operator.truediv


class GroupAsBinOpMod(GroupAsBinOp):
    """
    For BinOp with op = '*'
    """

    op = operator.mod


class GroupAsBinOpLeftShift(GroupAsBinOp):
    """
    For BinOp with op = '<<'
    """

    def binop(self, lhs, rhs):
        return to_int32(
            FoldedFragment(to_int32(lhs) << (to_uint32(rhs) & 0x1f), Number))


class GroupAsBinOpRightShift(GroupAsBinOp):
    """
    For BinOp with op = '>>'
    """

    def binop(self, lhs, rhs):
        return to_int32(
            FoldedFragment(to_int32(lhs) >> (to_uint32(rhs) & 0x1f), Number))


class GroupAsBinOpUnsignedRightShift(GroupAsBinOp):
    """
    For BinOp with op = '>>>'
    """

    def binop(self, lhs, rhs):
        return to_uint32(
            FoldedFragment(to_uint32(lhs) >> (to_uint32(rhs) & 0x1f), Number))


class GroupAsBinOpBitwise(GroupAsBinOp):
    """
    Generic binop bitwise
    """

    def binop(self, lhs, rhs):
        """
        The plus operator is a bit different from the rest.
        """

        return FoldedFragment(self.op(to_int32(lhs), to_int32(rhs)), Number)

    def op(self, lhs, rhs):
        raise NotImplementedError


class GroupAsBinOpBitwiseAnd(GroupAsBinOpBitwise):
    """
    For BinOp with op = '&'
    """

    op = operator.and_


class GroupAsBinOpBitwiseXor(GroupAsBinOpBitwise):
    """
    For BinOp with op = '^'
    """

    op = operator.xor


class GroupAsBinOpBitwiseOr(GroupAsBinOpBitwise):
    """
    For BinOp with op = '|'
    """

    op = operator.or_


class GroupAsBinOpLogicalAnd(GroupAsBinOp):

    def binop(self, lhs, rhs):
        if not to_boolean(lhs):
            return FoldedFragment(lhs.value, type(lhs.node))
        else:
            return FoldedFragment(rhs.value, type(rhs.node))


class GroupAsBinOpLogicalOr(GroupAsBinOp):

    def binop(self, lhs, rhs):
        if to_boolean(lhs):
            return FoldedFragment(lhs.value, type(lhs.node))
        else:
            return FoldedFragment(rhs.value, type(rhs.node))


class UnaryOptionalSpace(Token):

    def __call__(self, walk, dispatcher, node):
        value = '' if node.op in {'++', '--', '!', '~', '+', '-'} else ' '
        yield next(dispatcher.token(None, node, value, None))


class GroupAsUnaryExpr(Attr):

    def __call__(self, walk, dispatcher, node):
        if not isinstance(node, UnaryExpr):
            raise TypeError(
                "Ruletype token %r expects a 'UnaryExpr', got %r" % (
                    type(self).__name__, type(node).__name__))

        value = next(walk(dispatcher, node.value, definition=None))
        yield next(dispatcher.token(None, node, self.unaryop(value), None))

    def unaryop(self, value):
        # assumes to be ExtractedFragments
        value = to_number(value)
        if value == 'NaN':
            return FoldedFragment('NaN', Number)
        else:
            return FoldedFragment(self.op(value), Number)

    def op(self, value):
        raise NotImplementedError


class GroupAsUnaryExprPlus(GroupAsUnaryExpr):
    """
    For UnaryExpr with op = '+'
    """

    def op(self, value):
        return value


class GroupAsUnaryExprMinus(GroupAsUnaryExpr):
    """
    For UnaryExpr with op = '-'
    """

    def op(self, value):
        return - value


class GroupAsUnaryExprBitwiseNot(GroupAsUnaryExpr):
    """
    For UnaryExpr with op = '~'
    """

    def unaryop(self, value):
        return FoldedFragment(~ to_int32(value), Number)


class GroupAsUnaryExprLogicalNot(GroupAsUnaryExpr):
    """
    For UnaryExpr with op = '!'
    """

    def unaryop(self, value):
        return FoldedFragment(not to_boolean(value), Boolean)


class AttrSink(Attr):
    """
    Used to consume everything declared in the Attr.

    This is used to adapt the Deferrable types such that they get
    processed, but no elements are actually yielded.
    """

    def __call__(self, walk, dispatcher, node):
        list(super(AttrSink, self).__call__(walk, dispatcher, node))
        return
        yield  # pragma: no cover


class LiteralEval(Attr):
    """
    Assume the handler will produce a chunk of type string, and use
    literal_eval to turn it into some underlying value; current intended
    usage is for strings and numbers.
    """

    def __call__(self, walk, dispatcher, node):
        value = self._getattr(dispatcher, node)
        for chunk in walk(dispatcher, value, token=self):
            yield next(dispatcher.token(
                None, node, literal_eval(chunk.value), None))


class Raw(Token):
    """
    Simply yield the raw value as required.
    """

    def __call__(self, walk, dispatcher, node):
        yield next(dispatcher.token(None, node, self.value, None))


class OpString(Token):
    """
    Special case for the BinOp/UnaryExpr, to ensure that the raw value
    that was parsed into the ast node is returned instead of being
    treated, such that the original expression have their quotes
    preserved.
    """

    def __call__(self, walk, dispatcher, node):
        value = getattr(node, self.attr)
        if isinstance(value, String):
            yield next(dispatcher.token(None, node, value.value, None))
        else:
            for chunk in walk(dispatcher, value, token=self):
                yield chunk


class RawBoolean(Attr):
    """
    Simple yield the raw boolean value from the attribute.
    """

    def __call__(self, walk, dispatcher, node):
        value = self._getattr(dispatcher, node)
        if value == 'true':
            yield next(dispatcher.token(None, node, True, None))
        elif value == 'false':
            yield next(dispatcher.token(None, node, False, None))
        else:
            raise ValueError('%r is not a JavaScript boolean value' % value)


class TopLevelAttrs(Attr):
    """
    Denotes a top level attribute generator; should ensure all yielded
    values are all in the form of Assignment, otherwise collate them into
    the "default" NotImplemented key.

    Do note that by yielding naked `Assignment` type, in effect finalizes
    the walker as this type is not handled by other Attr types in this
    module.
    """

    def __call__(self, walk, dispatcher, node):
        misc_chunks = defaultdict(list)
        nodes = iter(node)
        for target_node in nodes:
            for chunk in walk(dispatcher, target_node, token=self):
                if isinstance(chunk, ExtractedFragment):
                    if isinstance(chunk.value, AssignmentList):
                        for assignment in chunk.value:
                            yield assignment
                    else:
                        misc_chunks[nodetype(chunk.node)].append(chunk.value)
                else:
                    # ideally, walk.throw() be called instead as the
                    # exception would propagate to the real ruletype
                    # responsible, but would also completely kill this
                    # generator; so instead just invoke the dispatcher
                    # error_handler.
                    dispatcher.error_handler(
                        TypeError(
                            "generated value %r is not an instance of "
                            "ExtractedFragment, thus it cannot be yielded by "
                            "instances of %r; check that all ruletypes "
                            "specified for target node type %r such that "
                            "the values generate by them are done through "
                            "walk or dispatcher.token, or yield instances of "
                            "ExtractedFragment." % (
                                chunk,
                                type(self),
                                target_node,
                            )
                        ),
                        rule=self,
                        node=target_node,
                    )

        if misc_chunks:
            for key, value in misc_chunks.items():
                yield Assignment(key, value)


value = (
    Attr('value'),
)
values = (
    JoinAttr(Iter()),
)
top_level_values = (
    TopLevelAttrs(),
)

# definitions of all the rules for all the types for an ES5 program.
definitions = {
    'ES5Program': top_level_values,
    'Block': (
        GroupAsMap((
            JoinAttr(Iter()),
        )),
    ),
    'VarStatement': values,
    'VarDecl': (
        AttrListAssignment(['identifier', 'initializer']),
    ),
    'VarDeclNoIn': (
        AttrListAssignment(['identifier', 'initializer']),
    ),
    'GroupingOp': (
        Attr('expr'),
    ),
    'Identifier': (
        Attr(Resolve()),
    ),
    'PropIdentifier': value,
    'Assign': (
        AttrListAssignment(['left', 'right']),
    ),
    'GetPropAssign': (
        GroupAsList((
            Attr('prop_name'),
            PushScope,
            GroupAsMap((
                JoinAttr(attr='elements'),
            )),
            PopScope,
        ),),
    ),
    'SetPropAssign': (
        GroupAsList((
            Attr('prop_name'),
            PushScope,
            GroupAsMap((
                JoinAttr(attr='elements'),
            )),
            PopScope,
        ),),
    ),
    'Number': (
        LiteralEval('value'),
    ),
    'Comma': (
        Attr('left'), Attr('right'),
    ),
    'EmptyStatement': (),
    'If': (
        GroupAsList((
            Attr('predicate'),
            GroupAsMap((
                Attr('consequent'),
            ),),
            GroupAsMap((
                Attr('alternative'),
            ),),
        ),),
    ),
    'Boolean': (
        RawBoolean('value'),
    ),
    'For': (
        Attr('init'),
        GroupAsList((
            Attr('cond'),
            Attr('count'),
            GroupAsMap((
                Attr('statement'),
            ),),
        )),
    ),
    'ForIn': (
        GroupAsList((
            Attr('item'),
            Attr('iterable'),
            GroupAsMap((
                Attr('statement'),
            ),),
        ),),
    ),
    'BinOp': (GroupAsStr((
        OpString('left'), Text(value=' '),
        Operator(attr='op'),
        Text(value=' '), OpString('right'),
    ),),),
    'UnaryExpr': (
        # the default grouping required to ensure positive/negative
        # numbers return as expected.
        OpDisambiguate(
            value={
                NotImplemented: (GroupAsStr(
                    (Attr('op'), UnaryOptionalSpace(), OpString('value')),
                    value='',
                ),),
                '+': (GroupAsUnaryExprPlus(),),
                '-': (GroupAsUnaryExprMinus(),),
            }
        ),
    ),
    'PostfixExpr': (
        Operator(attr='value'),
    ),
    'ExprStatement': (
        Attr('expr'),
    ),
    'DoWhile': (
        GroupAsList((
            GroupAsMap((Attr('statement'),),),
            Attr('predicate'),
        ),),
    ),
    'While': (
        GroupAsList((
            Attr('predicate'),
            GroupAsMap((Attr('statement'),),),
        ),),
    ),
    'Null': (
        Raw(value=None),
    ),
    'String': (
        LiteralEval(Literal()),
    ),
    'Continue': (),
    'Break': (),
    'Return': (
        Optional('expr', (
            GroupAsAssignment((
                Text(value='return'),
                Attr(attr='expr'),
            )),
        ),),
    ),
    'With': (
        GroupAsList((
            Attr('expr'),
            GroupAsMap((Attr('statement'),),),
        ),),
    ),
    'Label': (
        GroupAsAssignment((
            Attr(attr='identifier'),
            GroupAsMap((
                Attr(attr='statement'),
            )),
        )),
    ),
    'Switch': (
        GroupAsList((
            Attr('expr'),
            Attr('case_block'),
        ),),
    ),
    'CaseBlock': (
        GroupAsMap((
            JoinAttr(Iter()),
        )),
    ),
    'Case': (
        GroupAsList((
            Attr('expr'),
            GroupAsMap((
                JoinAttr('elements'),
            ),),
        ),),
    ),
    'Default': (
        GroupAsList((
            GroupAsMap((
                JoinAttr('elements'),
            ),),
        ),),
    ),
    'Throw': (
        Attr('expr'),
    ),
    'Debugger': (),
    'Try': (
        GroupAsList((
            Attr('statements'),
            Attr('catch'),
            Attr('fin'),
        ),),
    ),
    'Catch': (
        GroupAsMap((
            GroupAsList((
                Attr('identifier'),
                Attr('elements'),
            ),),
        ),),
    ),
    'Finally': (
        GroupAsMap((
            GroupAsList((
                Attr('elements'),
            ),),
        ),),
    ),
    'FuncDecl': (
        GroupAsAssignment((
            Attr(Declare('identifier')),
            PushScope,
            GroupAsList((
                Optional('identifier', (ResolveFuncName,)),
                GroupAsList((
                    JoinAttr(Declare('parameters'),),
                )),
                GroupAsMap((
                    JoinAttr('elements'),
                )),
            ),),
            PopScope,
        ),),
    ),
    'FuncExpr': (
        AttrSink(Declare('identifier')),
        PushScope,
        GroupAsList((
            Optional('identifier', (ResolveFuncName,)),
            GroupAsList((
                JoinAttr(Declare('parameters'),),
            )),
            GroupAsMap((
                JoinAttr('elements'),
            )),
        ),),
        PopScope,
    ),
    'Conditional': (
        GroupAsList((
            Attr('predicate'),
            Attr('consequent'),
            Attr('alternative'),
        )),
    ),
    'Regex': value,
    'NewExpr': (
        GroupAsList((Attr('identifier'), Attr('args'),),),
    ),
    'DotAccessor': (
        # The current way may simply result in a binding that has a dot,
        # this may be desirable now, however an alternative manner is to
        # implement registration and update of the value within the
        # scope...
        GroupAsStr(
            (Attr('node'), Text(value='.'), Attr('identifier'),),
            value='',
        ),
    ),
    'BracketAccessor': (
        # Likewise similar as above
        GroupAsStr(
            (Attr('node'), Text(value='['), Attr('expr'), Text(value=']'),),
            value='',
        ),
    ),
    'FunctionCall': (
        GroupAsList((Attr('identifier'), Attr('args'),),),
    ),
    'Arguments': (
        GroupAsList((JoinAttr('items',),),),
    ),
    'Object': (
        GroupAsMap((JoinAttr('properties'),)),
    ),
    'Array': (
        GroupAsList((JoinAttr('items'),),),
    ),
    'Elision': (),
    'This': (
        Text(value='this'),
    ),
    'Comments': (),
    'LineComment': (),
    'BlockComment': (),
}


def token_handler_extractor(
        token, dispatcher, node, subnode, sourcepath_stack=None):
    """
    The token handler that will yield the ExtractedFragment type.
    """

    # Ideally, some kind of simplified bytecode should be generated to
    # instruct how things are assigned by the above rules, such that the
    # origin node can then be encoded as part of the process, but this
    # is a rather huge pain to do (approaching the work of building a
    # proper language interpreter/bytecode compiler, so this shortcut is
    # taken instead.
    if isinstance(subnode, FoldedFragment):
        value, folded_type = subnode
    else:
        value, folded_type = subnode, nodetype(node)
    yield ExtractedFragment(value, node, folded_type)


class Dispatcher(walker.Dispatcher):

    def error_handler(self, exception, rule=None, node=None):
        logger.error(
            "failed to process node %r with rule %r to an extracted value; "
            "cause: %s: %s",
            node, type(rule).__name__, type(exception).__name__, exception
        )
        return next(self.token(rule, node, '', []))


class Unparser(BaseUnparser):

    def __init__(
            self,
            definitions=definitions,
            token_handler=token_handler_extractor,
            rules=(),
            layout_handlers=None,
            deferrable_handlers=None,
            prewalk_hooks=(),
            dispatcher_cls=walker.Dispatcher):

        super(Unparser, self).__init__(
            definitions=definitions,
            token_handler=token_handler_extractor,
            rules=rules,
            layout_handlers=layout_handlers,
            deferrable_handlers=deferrable_handlers,
            prewalk_hooks=prewalk_hooks,
            dispatcher_cls=dispatcher_cls,
        )


def extractor(fold_ops=False, ignore_errors=False):
    """
    A helper to construct an instance of the extractor Unparser

    arguments

    fold_ops
        Apply an update to a clone of the definitions defined at the
        root of this module such that the definition fo BinOp will be
        modified with a default set of GroupAsBinOp tokens that will
        process the operand with the associated operator token.

        This would result in constants (such as Numbers and Strings)
        being evaluated with the associated operator, yielding an
        evaluated expression or concatenated strings.  For other
        identifiers and types encountered, they are simply replaced with
        a string format replacement fields and are simply treated as if
        they are strings (i.e. result will be a concatenation).

        This flag has significant limitations and the resulting values
        of limited value, given that no JavaScript engine or real
        execution of the code is done.  It WILL produce wild and
        unexpected results.  Use with caution.

        Default: false (operators will simply be concatenated together
        with the operands as a string to produce values).

    ignore_errors
        Attempt to continue through errors triggered through processing
        of the input ast through the unparser.

        Default: false (exceptions will be thrown on error).
    """

    dispatcher_cls = Dispatcher if ignore_errors else walker.Dispatcher
    new_definitions = {}
    new_definitions.update(definitions)
    if fold_ops:
        new_definitions.update({
            'BinOp': (
                OpDisambiguate(
                    value={
                        NotImplemented: (GroupAsStr((
                            Attr('left'), Text(value=' '),
                            Operator(attr='op'),
                            Text(value=' '), Attr('right'),
                        ),),),
                        '+': (GroupAsBinOpPlus(),),
                        '-': (GroupAsBinOpMinus(),),
                        '*': (GroupAsBinOpMult(),),
                        '/': (GroupAsBinOpDiv(),),
                        '%': (GroupAsBinOpMod(),),
                        '<<': (GroupAsBinOpLeftShift(),),
                        '>>': (GroupAsBinOpRightShift(),),
                        '>>>': (GroupAsBinOpUnsignedRightShift(),),
                        '&': (GroupAsBinOpBitwiseAnd(),),
                        '^': (GroupAsBinOpBitwiseXor(),),
                        '|': (GroupAsBinOpBitwiseOr(),),
                        '&&': (GroupAsBinOpLogicalAnd(),),
                        '||': (GroupAsBinOpLogicalOr(),),
                    }
                ),
            ),
            'UnaryExpr': (
                OpDisambiguate(
                    value={
                        NotImplemented: (GroupAsStr(
                            (Attr('op'), UnaryOptionalSpace(),
                                OpString('value')),
                            value='',
                        ),),
                        '+': (GroupAsUnaryExprPlus(),),
                        '-': (GroupAsUnaryExprMinus(),),
                        '~': (GroupAsUnaryExprBitwiseNot(),),
                        '!': (GroupAsUnaryExprLogicalNot(),),
                    }
                ),
            ),
        })

    return Unparser(definitions=new_definitions, dispatcher_cls=dispatcher_cls)


def ast_to_dict(ast, fold_ops=False, ignore_errors=False):
    """
    Simple dictionary building function - return a dictionary for the
    source abstract syntax tree for the parsed program.

    arguments

    ast
        The AST to convert to a dictionary.

    fold_ops
        Apply an update to a clone of the definitions defined at the
        root of this module such that the definition fo BinOp will be
        modified with a default set of GroupAsBinOp tokens that will
        process the operand with the associated operator token.

        This would result in constants (such as Numbers and Strings)
        being evaluated with the associated operator, yielding an
        evaluated expression or concatenated strings.  For other
        identifiers and types encountered, they are simply replaced with
        a string format replacement fields and are simply treated as if
        they are strings (i.e. result will be a concatenation).

        This flag has significant limitations and the resulting values
        of limited value, given that no JavaScript engine or real
        execution of the code is done.  It WILL produce wild and
        unexpected results.  Use with caution.

        Default: false (operators will simply be concatenated together
        with the operands as a string to produce values).

    ignore_errors
        Attempt to continue through errors triggered through processing
        of the input ast through the unparser.

        Default: false (exceptions will be thrown on error).
    """

    return dict(extractor(fold_ops=fold_ops, ignore_errors=ignore_errors)(ast))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import textwrap
import unittest

from calmjs.parse.asttypes import (
    Array,
    Block,
    Boolean,
    Case,
    Catch,
    Default,
    DoWhile,
    Finally,
    For,
    ForIn,
    FunctionCall,
    Identifier,
    If,
    Object,
    Null,
    Number,
    GetPropAssign,
    SetPropAssign,
    String,
    Switch,
    Try,
    UnaryExpr,
    With,
    While,
)
from calmjs.parse.parsers import es5
from calmjs.parse.ruletypes import (
    Attr,
    JoinAttr,
)
from calmjs.parse.unparsers.extractor import (
    Assignment,
    AssignmentList,
    Dispatcher,
    ExtractedFragment,
    FoldedFragment,
    GroupAsBinOp,
    GroupAsBinOpBitwise,
    GroupAsBinOpPlus,
    GroupAsUnaryExpr,
    GroupAsUnaryExprPlus,
    Unparser,
    definitions,
    extractor,
    ast_to_dict,
    logger as extractor_logger,
    to_boolean,
    to_number,
    to_int32,
    to_primitive,
    to_string,
    value_to_str,
)
from calmjs.parse.testing.util import setup_logger


def parse(src):
    return es5.parse(textwrap.dedent(src).strip())


class SupportTestCase(unittest.TestCase):

    def test_assignment_repr(self):
        assignment = Assignment(
            ExtractedFragment(1, Number(value=1), None),
            ExtractedFragment(1, Number(value=1), None),
        )
        self.assertEqual('(1, 1)', repr(assignment))

        assignmentlist = AssignmentList(assignment)
        self.assertEqual('[(1, 1)]', repr(assignmentlist))

    def test_assignment_attributes(self):
        assignment = Assignment(1, 2)
        self.assertEqual(assignment.key, 1)
        self.assertEqual(assignment.value, 2)

        assignment_frag = Assignment(
            ExtractedFragment(1, Number(value=1), None),
            ExtractedFragment(2, Number(value=2), None),
        )
        self.assertEqual(assignment_frag.key, 1)
        self.assertEqual(assignment_frag.value, 2)

    def test_assignment_list_creation(self):
        assignmentlist = AssignmentList([1, 2])
        self.assertEqual(Assignment(1, 2), assignmentlist[0])
        self.assertEqual(1, len(assignmentlist))
        assignmentlist2 = AssignmentList(assignmentlist)
        self.assertIs(assignmentlist[0], assignmentlist2[0])

        with self.assertRaises(ValueError) as e:
            AssignmentList([1, 2, 3])

        with self.assertRaises(ValueError) as e:
            AssignmentList([1])

        self.assertEqual(
            e.exception.args[0],
            '[1] cannot be converted to an Assignment',
        )

    def test_assignment_list_modify(self):
        assignmentlist = AssignmentList([1, 2], [3, 4])
        assignmentlist[1] = [4, 5]
        self.assertEqual(Assignment(4, 5), assignmentlist[1])
        del assignmentlist[0]
        self.assertEqual(Assignment(4, 5), assignmentlist[0])
        self.assertEqual(1, len(assignmentlist))

    def test_assignment_list_generic_equality(self):
        assignmentlist = AssignmentList(Assignment(1, 2), Assignment(3, 4))
        self.assertEqual([(1, 2), (3, 4)], assignmentlist)


class TypeConversionTestCase(unittest.TestCase):

    def test_value_to_str(self):
        with self.assertRaises(TypeError):
            value_to_str(object())

    def test_to_primitive_various_lists(self):
        def assertArrayToPrimitive(hint, left, right):
            # local test helper; move out when appropriate.
            self.assertEqual(
                FoldedFragment(left, hint),
                to_primitive(FoldedFragment(right, Array), hint),
            )

        assertArrayToPrimitive(String, '', [])
        assertArrayToPrimitive(String, '1', [1])
        assertArrayToPrimitive(String, '1,2', [1, 2])
        assertArrayToPrimitive(String, '1,2,3', [1, 2, 3])
        assertArrayToPrimitive(
            String, '1,2,3,4,5,6,7,8', [1, 2, [3, 4], 5, [[[6], 7], 8]])
        assertArrayToPrimitive(
            String, '[object Object],[object Object]', [{}, {}])
        assertArrayToPrimitive(
            String,
            '0,1,,true,false,hello', [0, 1, None, True, False, 'hello'])

        assertArrayToPrimitive(Number, 0, [])
        assertArrayToPrimitive(Number, 0, [0])
        assertArrayToPrimitive(Number, 1, [1])
        assertArrayToPrimitive(Number, 1, ['1'])
        assertArrayToPrimitive(Number, 'NaN', [1, 2])
        assertArrayToPrimitive(Number, 1, [[[1]]])

        assertArrayToPrimitive(Number, 'NaN', [True])
        assertArrayToPrimitive(Number, 'NaN', [False])
        assertArrayToPrimitive(Number, 2748, ['  0002748 '])
        assertArrayToPrimitive(Number, 2748, ['  0xabc  '])
        assertArrayToPrimitive(Number, 'NaN', ['  00xabc  '])

    def test_to_primitive_object(self):
        # TODO validate that all the folded_type returned is expected
        self.assertEqual('[object Object]', to_primitive(
            FoldedFragment({}, Object), String).value)
        self.assertEqual('NaN', to_primitive(
            FoldedFragment({}, Object), Number).value)

    def test_to_primitive_other(self):
        # TODO validate that all the folded_type returned is expected
        fragments = [
            FoldedFragment(0, Number),
            FoldedFragment('hello', String),
            FoldedFragment(True, Boolean),
            FoldedFragment(None, Null),
        ]
        for fragment in fragments:
            self.assertIs(fragment, to_primitive(fragment, String))
        for fragment in fragments:
            self.assertIs(fragment, to_primitive(fragment, Number))

    def test_to_boolean(self):
        self.assertTrue(to_boolean(FoldedFragment([], Array)))
        self.assertTrue(to_boolean(FoldedFragment({}, Object)))
        self.assertFalse(to_boolean(FoldedFragment(None, Null)))
        self.assertTrue(to_boolean(FoldedFragment(True, Boolean)))
        self.assertFalse(to_boolean(FoldedFragment(False, Boolean)))
        self.assertTrue(to_boolean(FoldedFragment(1, Number)))
        self.assertFalse(to_boolean(FoldedFragment(0, Number)))
        self.assertFalse(to_boolean(FoldedFragment('NaN', Number)))
        self.assertTrue(to_boolean(FoldedFragment('hello', String)))
        self.assertFalse(to_boolean(FoldedFragment('', String)))

    def test_to_number(self):
        self.assertEqual(0, to_number(FoldedFragment([], Array)))
        self.assertEqual(1, to_number(FoldedFragment([[['1']]], Array)))
        self.assertEqual(1, to_number(FoldedFragment(True, Boolean)))
        self.assertEqual(0, to_number(FoldedFragment(False, Boolean)))
        value = 5343
        self.assertIs(value, to_number(FoldedFragment(value, Number)))
        self.assertEqual(0, to_number(FoldedFragment(None, Null)))
        self.assertEqual(1234, to_number(FoldedFragment('1234', Array)))
        self.assertEqual('NaN', to_number(FoldedFragment({}, Object)))

        # for completeness, no idea how might other asttypes slip through
        self.assertEqual('NaN', to_number(FoldedFragment([], Block)))

    def test_to_string(self):
        self.assertEqual('', to_string(FoldedFragment([], Array)))
        self.assertEqual('true', to_string(FoldedFragment(True, Boolean)))
        self.assertEqual('false', to_string(FoldedFragment(False, Boolean)))
        self.assertEqual('null', to_string(FoldedFragment(None, Null)))
        self.assertEqual(
            '12345.67', to_string(FoldedFragment(12345.67, Number)))
        value = 'hello world'
        self.assertIs(value, to_string(FoldedFragment(value, String)))
        self.assertEqual(
            '[object Object]', to_string(FoldedFragment({}, Object)))

        # for completeness, there could be definitions where unsupported
        # asttypes can slip through, just ensure they output an empty
        # string if the produced value is not a string.
        self.assertEqual('', to_string(FoldedFragment([], Block)))
        # for string values, enclose them as a format token.
        self.assertEqual(
            '{value}', to_string(FoldedFragment('value', Identifier)))

    def test_to_int32(self):
        self.assertEqual(0, to_int32(FoldedFragment([], Array)))
        self.assertEqual(1, to_int32(FoldedFragment([[['1']]], Array)))
        self.assertEqual(1, to_int32(FoldedFragment(True, Boolean)))
        self.assertEqual(0, to_int32(FoldedFragment(False, Boolean)))

        self.assertEqual(54321, to_int32(FoldedFragment(54321, Number)))

        self.assertEqual(0, to_int32(FoldedFragment(None, Null)))
        self.assertEqual(1234, to_int32(FoldedFragment('1234', Array)))
        self.assertEqual(0, to_int32(FoldedFragment({}, Object)))

        maxint32 = 2147483647
        overint32 = 2147483648
        self.assertEqual(
            maxint32, to_int32(FoldedFragment(maxint32, Number)))
        self.assertEqual(
            - overint32, to_int32(FoldedFragment(overint32, Number)))

        # for completeness, no idea how might other asttypes slip through
        self.assertEqual(0, to_int32(FoldedFragment([], Block)))


class ExtractorUnparserErrorTestCase(unittest.TestCase):

    def test_bad_boolean(self):
        unparser = Unparser()
        ast = parse('thing = false;')
        # mangle the type such that the lookup fails
        ast.children()[0].expr.right.value = 'none'
        with self.assertRaises(ValueError) as e:
            list(unparser(ast))
        self.assertIn(
            "'none' is not a JavaScript boolean value",
            e.exception.args[0],
        )

    def assert_definitions_fault(
            self, errorcls, definitions, src, msg, result):
        unparser = Unparser(definitions=definitions)
        ast = parse(src)
        with self.assertRaises(errorcls) as e:
            dict(unparser(ast))
        self.assertIn(msg, str(e.exception))

        err = setup_logger(self, extractor_logger)
        unparser = Unparser(definitions=definitions, dispatcher_cls=Dispatcher)
        ast = parse(src)
        self.assertEqual(result, dict(unparser(ast)))
        self.assertIn(msg, err.getvalue())

    def test_definitions_fault_handling(self):
        class BadAttr(Attr):
            def __call__(self, walk, dispatcher, node):
                # rather than going through dispatcher.walk, yield this
                # unwrapped type (i.e. not via walk or dispatcher.token
                yield self._getattr(dispatcher, node)

        mismatched = {}
        mismatched.update(definitions)
        mismatched['Assign'] = (Attr('left'), BadAttr('op'), Attr('right'),)

        self.assert_definitions_fault(
            TypeError, mismatched,
            'a = 1;{}',
            "'=' is not an instance of ExtractedFragment",
            {
                Identifier: ['a'],
                # Assign: ['='],
                Number: [1],
                # sentinel node to ensure no other premature termination
                Block: [{}],
            },
        )

    def test_definitions_groupasunary_errors(self):
        # for the case where certain types have not been properly
        # reduced to a supported grouping
        faulty = {}
        faulty.update(definitions)
        faulty['Array'] = (GroupAsUnaryExprPlus(),)
        faulty['UnaryExpr'] = (GroupAsUnaryExpr(),)

        self.assert_definitions_fault(
            TypeError, faulty,
            "a = [1, 2, 3]",
            "Ruletype token 'GroupAsUnaryExprPlus' expects a 'UnaryExpr', "
            "got 'Array'",
            {'a': ''},
        )
        self.assert_definitions_fault(
            NotImplementedError, faulty,
            "a = +1",
            "",
            {'a': ''},
        )

    def test_definitions_groupasbin_errors(self):
        # for the case where certain types have not been properly
        # reduced to a supported grouping
        faulty = {}
        faulty.update(definitions)
        faulty['Array'] = (GroupAsBinOpPlus(),)

        self.assert_definitions_fault(
            TypeError, faulty,
            "a = [1, 2, 3]",
            "Ruletype token 'GroupAsBinOpPlus' expects a 'BinOp', got 'Array'",
            {'a': ''},
        )

        faulty['BinOp'] = (GroupAsBinOp(),)
        self.assert_definitions_fault(
            NotImplementedError, faulty,
            "a = 1 + 1; b = 'x' + 'x'",
            "",
            {'a': '', 'b': 'NaN'},
        )

        faulty['BinOp'] = (GroupAsBinOpBitwise(),)
        self.assert_definitions_fault(
            NotImplementedError, faulty,
            "a = 1 | 1; b = 'x' | 'x'",
            "",
            {'a': '', 'b': ''},
        )

    def test_definitions_groupasbin_incompat(self):
        # for the case where certain types have not been properly
        # reduced to a supported grouping
        faulty = {}
        faulty.update(definitions)
        faulty['Array'] = (JoinAttr('items'),)
        faulty['BinOp'] = (GroupAsBinOpPlus(),)

        self.assert_definitions_fault(
            ValueError, faulty,
            "a = '' + [1, 2, 3]",
            "Ruletype token 'GroupAsBinOpPlus' unable to process output "
            "produced by the definition used on <Array @1:10 ...>, as it "
            "yielded more than one fragment (first two fragments are "
            "ExtractedFragment(value=1, node=<Number @1:11 value='1'>, "
            "folded_type=<class 'calmjs.parse.asttypes.Number'>) and "
            "ExtractedFragment(value=2, node=<Number @1:14 value='2'>, "
            "folded_type=<class 'calmjs.parse.asttypes.Number'>))",
            {'a': ''},
        )


class ExtractorUnparserTestCase(unittest.TestCase):

    def test_empty_program(self):
        unparser = Unparser()
        ast = parse('')
        self.assertEqual(dict(unparser(ast)), {})

    def test_empty_statements(self):
        unparser = Unparser()
        ast = parse(";;;")
        self.assertEqual(dict(unparser(ast)), {})

    def test_ignoreed_statements(self):
        unparser = Unparser()
        ast = parse("""
        ;
        debugger;
        continue;
        break;
        """)
        self.assertEqual(dict(unparser(ast)), {})

    def test_singular_atom(self):
        unparser = Unparser()
        ast = parse('0;')
        self.assertEqual(dict(unparser(ast)), {
            Number: [0],
        })

    def test_multiple_atoms(self):
        unparser = Unparser()
        ast = parse('{}0;{}1;{}2;')
        self.assertEqual(dict(unparser(ast)), {
            Number: [0, 1, 2],
            Block: [{}, {}, {}],
        })

    def test_singular_assignment_pair(self):
        unparser = Unparser()
        ast = parse('''
        var x = 0;
        ''')
        self.assertEqual(dict(unparser(ast)), {
            'x': 0,
        })

    def test_inline_double(self):
        unparser = Unparser()
        ast = parse('''
        a = b = 'hello';
        var x = y = 42;
        ''')
        self.assertEqual(dict(unparser(ast)), {
            'a': 'hello',
            'b': 'hello',
            'x': 42,
            'y': 42,
        })

    def test_var_multiple(self):
        unparser = Unparser()
        ast = parse('''
        var x = 69, y = 420;
        ''')
        self.assertEqual(dict(unparser(ast)), {
            'x': 69,
            'y': 420,
        })

    def test_var_undefined(self):
        unparser = Unparser()
        ast = parse('''
        var x, y;
        ''')
        # JavaScript `undefined` values are simply casted to None
        self.assertEqual(dict(unparser(ast)), {
            'x': None,
            'y': None,
        })

    def test_inline_multiassign(self):
        unparser = Unparser()
        ast = parse('''
        a = b = c = d = 'hello';
        var x = y = z = 42;
        var r = s = /abc/;
        ''')
        self.assertEqual(dict(unparser(ast)), {
            'a': 'hello',
            'b': 'hello',
            'c': 'hello',
            'd': 'hello',
            'x': 42,
            'y': 42,
            'z': 42,
            'r': '/abc/',
            's': '/abc/',
        })

    def test_multiple_assignment_pairs(self):
        unparser = Unparser()
        ast = parse('''
        var x = 0;
        var y = 'abc';
        ''')
        self.assertEqual(dict(unparser(ast)), {
            'x': 0,
            'y': 'abc',
        })

    def test_list_assignment(self):
        unparser = Unparser()
        ast = parse('''
        var a_list = [1, 2, 3, 'foobar', 0];
        var b_list = [];
        ''')
        self.assertEqual(dict(unparser(ast)), {
            'a_list': [1, 2, 3, 'foobar', 0],
            'b_list': [],
        })

    def test_list_in_a_list(self):
        unparser = Unparser()
        ast = parse('''
        var a_list = [1, [2, [3, 'foobar', 0]]];
        ''')
        self.assertEqual(dict(unparser(ast)), {
            'a_list': [1, [2, [3, 'foobar', 0]]],
        })

    def test_object_assignment_basic(self):
        unparser = Unparser()
        ast = parse('''
        var obj_a = {
            a: 1,
            'b': 2,
            'c': [1, 2],
            d: /a/,
            e: /a/i,
        }
        ''')
        self.assertEqual(dict(unparser(ast)), {
            'obj_a': {'a': 1, 'b': 2, 'c': [1, 2], 'd': '/a/', 'e': '/a/i'}
        })

    def test_object_assignment_getter_setter(self):
        unparser = Unparser()
        ast = parse('''
        var obj_a = {
            foo: 1,
            get bar() {
                x = 1;
                return x;
            },
            set bar(x) {
                x = 2;
                return
            },
        }
        ''')
        # getter and setters are dumped out as an item under the
        # NotImplemented key.
        self.assertEqual(dict(unparser(ast)), {
            'obj_a': {
                'foo': 1,
                GetPropAssign: [
                    ['bar', {'x': 1, 'return': 'x'}],
                ],
                SetPropAssign: [
                    ['bar', {'x': 2}],
                ],
            },
        })

    def test_dot_accessor_assignment(self):
        unparser = Unparser()
        ast = parse('''
        some.attrib.a = 'value'
        ''')
        self.assertEqual(dict(unparser(ast)), {
            'some.attrib.a': 'value'
        })

    def test_bracket_accessor_assignment(self):
        unparser = Unparser()
        ast = parse('''
        some['attrib']['a'] = 'value'
        some[id]['a'] = 'value'
        ''')
        # Note that the quoted string is dropped to be an identifier
        self.assertEqual(dict(unparser(ast)), {
            'some[attrib][a]': 'value',
            'some[id][a]': 'value',
        })

    def test_bracket_accessor_on_string(self):
        unparser = Unparser()
        ast = parse('''
        value = 'hello'[1];
        ''')
        # bracket accessor won't try to resolve the value
        self.assertEqual(dict(unparser(ast)), {
            'value': 'hello[1]',
        })

    def test_various_special_values(self):
        unparser = Unparser()
        ast = parse('''
        a = true;
        b = false;
        c = null;
        ''')
        self.assertEqual(dict(unparser(ast)), {
            'a': True,
            'b': False,
            'c': None,
        })

    def test_basic_operator_str(self):
        unparser = Unparser()
        ast = parse("""
        a = 'hello' + ' ' + 'world';
        b = !'hello' + ' ' + 'world';
        """)
        self.assertEqual(dict(unparser(ast)), {
            'a': "'hello' + ' ' + 'world'",
            'b': "!'hello' + ' ' + 'world'",
        })

    def test_ternary_assignment(self):
        unparser = Unparser()
        ast = parse("""
        a = 1 ? '1' : false;
        """)
        self.assertEqual(dict(unparser(ast)), {
            'a': [1, '1', False],
        })

    def test_funcdecl(self):
        unparser = Unparser()
        ast = parse("""
        function a() {
            foo = 1;
        }
        function b() {
            bar = 2;
        }
        """)
        self.assertEqual(dict(unparser(ast)), {
            'a': [[], {
                'foo': 1,
            }],
            'b': [[], {
                'bar': 2,
            }],
        })

    def test_funcexpr_assignment(self):
        unparser = Unparser()
        ast = parse("""
        a = function a() {
            foo = 1;
        };
        """)
        self.assertEqual(dict(unparser(ast)), {
            'a': [[], {
                'foo': 1,
            }],
        })

    def test_funcexpr_argument_assignment(self):
        unparser = Unparser()
        ast = parse("""
        a = function a(arg1, arg2) {
            foo = 1;
        };
        """)
        self.assertEqual(dict(unparser(ast)), {
            'a': [['arg1', 'arg2'], {
                'foo': 1,
            }]
        })

    def test_functioncall(self):
        unparser = Unparser()
        ast = parse("""
        foo = bar(a, b, d);
        """)
        # The result is due the grouping operator and the nested nature
        # of the anonymous call.
        self.assertEqual(dict(unparser(ast)), {
            'foo': ['bar', ['a', 'b', 'd']],
        })

    def test_new_functioncall(self):
        unparser = Unparser()
        ast = parse("""
        foo = new bar(a, b, d);
        """)
        # The result is due the grouping operator and the nested nature
        # of the anonymous call.
        self.assertEqual(dict(unparser(ast)), {
            'foo': ['bar', ['a', 'b', 'd']],
        })

    def test_funcexpr_anonymous_functioncall(self):
        unparser = Unparser()
        ast = parse("""
        (function () {
            foo = 1;
        })();
        """)
        # The result is due the grouping operator and the nested nature
        # of the anonymous call.
        self.assertEqual(dict(unparser(ast)), {
            FunctionCall: [[[[], {'foo': 1}], []]]
        })

    def test_label(self):
        unparser = Unparser()
        ast = parse("""
        var foo;
        labela:
            foo = 'bar';
        bar = 'baz';
        """)
        self.assertEqual(dict(unparser(ast)), {
            'foo': None,
            'labela': {'foo': 'bar'},
            'bar': 'baz',
        })

    def test_for_basic(self):
        unparser = Unparser()
        ast = parse("""
        a = 0
        for (; a < 1;) {
          a = a + 1
        }
        for (;;) {
        }
        for (i = 0, j = 10; i < j && j < 15; i++, j++) {
          x = i * j;
        }
        """)
        self.assertEqual(dict(unparser(ast)), {
            For: [
                ['a < 1', {Block: [{'a': 'a + 1'}]}],
                [{Block: [{}]}],
                ['i < j && j < 15', 'i', 'j', {Block: [{'x': 'i * j'}]}],
            ],
            'a': 0,
            'i': 0,
            'j': 10,
        })

    def test_for_var(self):
        unparser = Unparser()
        # this includes Comma, PostfixExpr, UnaryExpr
        ast = parse("""
        for (var i = 0, j = 10; i < j && j < 15; i++, ++j) {
          x = i * j;
        }
        """)
        self.assertEqual(dict(unparser(ast)), {
            For: [
                ['i < j && j < 15', 'i', '++j', {Block: [{'x': 'i * j'}]}],
            ],
            'i': 0,
            'j': 10,
        })

    def test_for_in_various(self):
        unparser = Unparser()
        ast = parse("""
        for (index in [1,2,3]) index
        for (index in [1,2,3]) {
            i = index;
        }
        """)
        self.assertEqual(dict(unparser(ast)), {
            ForIn: [
                ['index', [1, 2, 3], {Identifier: ['index']}],
                ['index', [1, 2, 3], {Block: [{'i': 'index'}]}],
            ],
        })

    def test_if_statement_various(self):
        unparser = Unparser()
        ast = parse("""
        if (true)
            var x = 100;
        else
            var x = 200;

        if (x < 100) {
            y = 1;
            z = 2;
        }
        else if (x < 200) {
            x = 2;
            y = 4;
        }
        else {
            x = 0;
            y = 0;
        }
        """)
        self.assertEqual(dict(unparser(ast)), {
            If: [
                [True, {'x': 100}, {'x': 200}],
                ['x < 100', {
                    Block: [{'y': 1, 'z': 2}],
                }, {
                    If: [[
                        'x < 200',
                        {Block: [{'x': 2, 'y': 4}]},
                        {Block: [{'x': 0, 'y': 0}]},
                    ]]
                }]
            ]
        })

    def test_while(self):
        unparser = Unparser()
        ast = parse("""
        while (false) {
          x = 1;
          y = 2;
          continue;
        }

        while (true) {
          x = 2;
          y = 4;
          continue foo;
        }
        """)
        self.assertEqual(dict(unparser(ast)), {
            While: [
                [False, {Block: [{
                    'x': 1,
                    'y': 2,
                }]}],
                [True, {Block: [{
                    'x': 2,
                    'y': 4,
                }]}],
            ],
        })

    def test_do_while(self):
        unparser = Unparser()
        ast = parse("""
        do {
          x = 1;
          y = 2;
        } while (false);

        do {
          x = 2;
          y = 4;
        } while (true);
        """)
        self.assertEqual(dict(unparser(ast)), {
            DoWhile: [
                [{Block: [{
                    'x': 1,
                    'y': 2,
                }]}, False],
                [{Block: [{
                    'x': 2,
                    'y': 4,
                }]}, True],
            ],
        })

    def test_while_assign(self):
        unparser = Unparser()
        ast = parse("""
        while (x = x + 1) {
          y = 2;
          break;
        }
        """)
        self.assertEqual(dict(unparser(ast)), {
            While: [
                [[('x', 'x + 1')], {Block: [{
                    'y': 2,
                }]}],
            ],
        })

    def test_with_statement(self):
        unparser = Unparser()
        ast = parse("""
        with (x) {
          x = x * 2;
        }
        """)
        self.assertEqual(dict(unparser(ast)), {
            With: [
                ['x', {Block: [{
                    'x': 'x * 2',
                }]}],
            ],
        })

    def test_with_statement_with_assign(self):
        unparser = Unparser()
        ast = parse("""
        with (x = 3) {
          x = x * 2;
        }
        """)
        self.assertEqual(dict(unparser(ast)), {
            With: [
                [[('x', 3)], {Block: [{
                    'x': 'x * 2',
                }]}],
            ],
        })

    def test_switch_case_blocks(self):
        unparser = Unparser()
        ast = parse("""
        switch (result) {
          case 'good':
            gooder = 1;
          case 'poor':
            poorer = 2;
            break;
          default:
            unexpected = 3;
          case 'errored':
            error = 4;
        }
        """)
        self.assertEqual(dict(unparser(ast)), {
            Switch: [[
                'result', {
                    Case: [
                        ['good', {'gooder': 1}],
                        ['poor', {'poorer': 2}],
                        ['errored', {'error': 4}],
                    ],
                    Default: [
                        [{'unexpected': 3}],
                    ],
                },
            ]],
        })

    def test_try_catch_finally_statement(self):
        unparser = Unparser()
        ast = parse("""
        try {
            var x = 100;
            y = 111;
        }
        catch (e) {
            var x = 200;
            y = 222;
        }
        finally {
            z = 1;
        }
        """)
        self.assertEqual(dict(unparser(ast)), {
            Try: [[{
                'x': 100,
                'y': 111,
            }, {
                Catch: [['e', {
                    'x': 200,
                    'y': 222,
                }]],
            }, {
                Finally: [[{
                    'z': 1,
                }]],
            }]],
        })

    def test_try_catch_try_finally_statement(self):
        unparser = Unparser()
        ast = parse("""
        try {
            var x = 100;
            y = 111;
        }
        catch (e) {
            var x = 200;
            y = 222;
        }

        try {
            var x = 10;
            y = 11;
        }
        finally {
            z = 3;
        }
        """)
        self.assertEqual(dict(unparser(ast)), {
            Try: [[{
                'x': 100,
                'y': 111,
            }, {
                Catch: [['e', {
                    'x': 200,
                    'y': 222,
                }]],
            }], [{
                'x': 10,
                'y': 11,
            }, {
                Finally: [[{
                    'z': 3,
                }]],
            }]],
        })

    def test_unary_expr(self):
        unparser = Unparser()
        ast = parse("""
        ++obj;
        delete obj;
        +42;
        -42;  // no negative numbers in ast; unary `-` negates positives
        !"foo";
        """)
        self.assertEqual(dict(unparser(ast)), {
            # UnaryExpr instead of Number, because that's the most outer
            # node captured; a "neat" demonstration of a slippage in the
            # ECMAScript syntax and semantics in its specifications.
            UnaryExpr: ['++obj', 'delete obj', 42, -42, '!"foo"'],
        })

    def test_postfix_op(self):
        unparser = Unparser()
        ast = parse("""
        i++;
        j--;
        """)
        self.assertEqual(dict(unparser(ast)), {
            Identifier: ['i', 'j'],
        })

    def test_elision(self):
        unparser = Unparser()
        ast = parse("""
        i = [,,,]
        """)
        self.assertEqual(dict(unparser(ast)), {'i': []})


class ExtractorTestCase(unittest.TestCase):
    """
    Simply test via the simplified constructor
    """

    def test_extractor_empty(self):
        ast = parse('')
        self.assertEqual({}, dict(extractor()(ast)))

    def test_bad_boolean_skipped(self):
        err = setup_logger(self, extractor_logger)
        ast = parse('thing = false;')
        # mangle the type such that the lookup fails
        ast.children()[0].expr.right.value = 'none'
        # use the ast_to_dict function
        result = ast_to_dict(ast, ignore_errors=True)
        self.assertIn(
            "failed to process node <Boolean @1:9 value='none'> with rule "
            "'RawBoolean' to an extracted value; cause: ValueError: ",
            err.getvalue(),
        )
        self.assertIn(
            "'none' is not a JavaScript boolean value",
            err.getvalue(),
        )
        # default value current an empty string.
        self.assertEqual(result, {'thing': ''})

    def test_binop_plus(self):
        ast = parse("""
        greetings = 'Hello, ' + 'World!';
        three = 1 + 2;
        _x1 = 'x' + 1;
        _1x = 1 + 'x';
        six = 1 + 2 + 3;
        objobj = {} + {};
        arrarr = [] + [];
        str12 = [1] + [2];
        strarr = [1, 2, 3] + [4, 5, 6];
        """)
        result = ast_to_dict(ast, fold_ops=True)
        self.assertEqual(result['greetings'], 'Hello, World!')
        self.assertEqual(result['three'], 3)
        self.assertEqual(result['_x1'], 'x1')
        self.assertEqual(result['_1x'], '1x')
        self.assertEqual(result['six'], 6)
        self.assertEqual(result['objobj'], '[object Object][object Object]')
        self.assertEqual(result['arrarr'], '')
        self.assertEqual(result['str12'], '12')
        self.assertEqual(result['strarr'], '1,2,34,5,6')

    def test_binop_plus_format(self):
        ast = parse("""
        name = "John";
        greetings = 'Hello, ' + name + '!';
        accessors = 'Poking at obj ' + a.b.c + ':' + thing[1]['abc'].x + '.';
        f_call = 'Function call: ' + f(a, b, c);
        namename = name + name;
        """)
        result = ast_to_dict(ast, fold_ops=True)
        self.assertEqual(result['name'], 'John')
        self.assertEqual(result['greetings'], 'Hello, {name}!')
        self.assertEqual(
            result['accessors'], 'Poking at obj {a.b.c}:{thing[1][abc].x}.')
        self.assertEqual(
            result['f_call'], 'Function call: {}')
        self.assertEqual(result['namename'], '{name}{name}')

    def test_binop_minus(self):
        ast = parse("""
        greetings = 'Hello, ' - 'World!';
        neg1 = 1 - 2;
        _x1 = 'x' - 1;
        _1x = 1 - 'x';
        neg4 = 1 - 2 - 3;
        wat = 'x' - 'x';
        objobj = {} - {};
        arrarr = [] - [];
        arrneg1 = [1] - [2];
        id = id1 - id2;
        """)
        result = ast_to_dict(ast, fold_ops=True)
        self.assertEqual(result['greetings'], 'NaN')
        self.assertEqual(result['neg1'], -1)
        self.assertEqual(result['_x1'], 'NaN')
        self.assertEqual(result['_1x'], 'NaN')
        self.assertEqual(result['neg4'], -4)
        self.assertEqual(result['wat'], 'NaN')
        self.assertEqual(result['objobj'], 'NaN')
        self.assertEqual(result['arrarr'], 0)
        self.assertEqual(result['arrneg1'], -1)
        self.assertEqual(result['id'], 'NaN')

    def test_binop_mult(self):
        ast = parse("""
        twelve = 3 * 4;
        sixty = 3 * 4 * 5;
        wot = 'x' * 'x';
        wut = 'x' * 'x' * 'x';
        // currently not working because NaN is treated as an identifier
        // check = NaN * NaN
        // currently done via grouping test
        objobj = {} * {};
        arrarr = [] * [];
        arrmult = [1] * [2];
        id = id1 * id2;
        """)
        result = ast_to_dict(ast, fold_ops=True)
        self.assertEqual(result['twelve'], 12)
        self.assertEqual(result['sixty'], 60)
        self.assertEqual(result['wot'], 'NaN')
        self.assertEqual(result['wut'], 'NaN')
        self.assertEqual(result['objobj'], 'NaN')
        self.assertEqual(result['arrarr'], 0)
        self.assertEqual(result['arrmult'], 2)
        self.assertEqual(result['id'], 'NaN')

    def test_binop_div(self):
        ast = parse("""
        threequarters = 3 / 4;
        ten = 1000 / 10 / 10;
        wot = 'x' / 'x';
        wut = 'x' / 'x' / 'x';
        objobj = {} / {};
        arrarr = [] / [];
        half = [1] / [2];
        id = id1 / id2;
        """)
        result = ast_to_dict(ast, fold_ops=True)
        self.assertEqual(result['threequarters'], 0.75)
        self.assertEqual(result['ten'], 10)
        self.assertEqual(result['wot'], 'NaN')
        self.assertEqual(result['wut'], 'NaN')
        self.assertEqual(result['objobj'], 'NaN')
        self.assertEqual(result['arrarr'], 'NaN')
        self.assertEqual(result['half'], 0.5)
        self.assertEqual(result['id'], 'NaN')

    def test_binop_mod(self):
        ast = parse("""
        zero = 2 % 2;
        one = 4 % 3;
        two = 13 % 11 % 3
        wot = 'x' % 'x';
        wut = 'x' % 'x' % 'x';
        objobj = {} % {};
        arrarr = [] % [];
        three = [3] % [4];
        id = id1 % id2;
        """)
        result = ast_to_dict(ast, fold_ops=True)
        self.assertEqual(result['zero'], 0)
        self.assertEqual(result['one'], 1)
        self.assertEqual(result['two'], 2)
        self.assertEqual(result['wot'], 'NaN')
        self.assertEqual(result['wut'], 'NaN')
        self.assertEqual(result['objobj'], 'NaN')
        self.assertEqual(result['arrarr'], 'NaN')
        self.assertEqual(result['three'], 3)
        self.assertEqual(result['id'], 'NaN')

    def test_binop_leftshift(self):
        ast = parse("""
        eight = 2 << 2;
        eightyeight = 11 << 3;
        wot = 'x' << 'x';
        objobj = {} << {};
        arrarr = [] << [];
        six = [3] << [1];
        sixteen = '4' << '2';
        id = id1 << id2;
        neg2 = 2147483647 << 1;
        negmax = 1 << 31;
        negmax2 = 2147483647 << 31;
        negspot = -2147483649 << 111;
        negtwo = -1 << -31;
        negone = -1 << -32;
        negmax3 = -1 << -33;
        """)
        result = ast_to_dict(ast, fold_ops=True)
        self.assertEqual(result['eight'], 8)
        self.assertEqual(result['eightyeight'], 88)
        self.assertEqual(result['wot'], 0)
        self.assertEqual(result['objobj'], 0)
        self.assertEqual(result['arrarr'], 0)
        self.assertEqual(result['six'], 6)
        self.assertEqual(result['sixteen'], 16)
        self.assertEqual(result['id'], 0)
        self.assertEqual(result['neg2'], -2)
        self.assertEqual(result['negmax'], -2147483648)
        self.assertEqual(result['negmax2'], -2147483648)
        self.assertEqual(result['negspot'], -32768)
        self.assertEqual(result['negtwo'], -2)
        self.assertEqual(result['negone'], -1)
        self.assertEqual(result['negmax3'], -2147483648)

    def test_binop_rightshift(self):
        ast = parse("""
        eight = 32 >> 2;
        eightyeight = 704 >> 3;
        wot = 'x' >> 'x';
        objobj = {} >> {};
        arrarr = [] >> [];
        six = [12] >> [1];
        sixteen = '259' >> '4';
        id = id1 >> id2;
        neg2 = -4 >> 1;
        negmax = 2147483648 >> [];
        max = 2147483647 >> 32;
        zero = 32767 >> 18;
        """)
        result = ast_to_dict(ast, fold_ops=True)
        self.assertEqual(result['eight'], 8)
        self.assertEqual(result['eightyeight'], 88)
        self.assertEqual(result['wot'], 0)
        self.assertEqual(result['objobj'], 0)
        self.assertEqual(result['arrarr'], 0)
        self.assertEqual(result['six'], 6)
        self.assertEqual(result['sixteen'], 16)
        self.assertEqual(result['id'], 0)
        self.assertEqual(result['neg2'], -2)
        self.assertEqual(result['negmax'], -2147483648)
        self.assertEqual(result['max'], 2147483647)
        self.assertEqual(result['zero'], 0)

    def test_binop_signedrightshift(self):
        ast = parse("""
        eight = 32 >>> 2;
        eightyeight = 704 >>> 3;
        wot = 'x' >>> 'x';
        objobj = {} >>> {};
        arrarr = [] >>> [];
        six = [12] >>> [1];
        sixteen = '259' >>> '4';
        id = id1 >>> id2;
        neg2 = -4 >>> 1;
        negmax = 2147483648 >>> [];
        max = -1 >>> 0;
        zero = 4294967296 >>> 0;
        overflow = 32767 >>> 18;
        """)
        result = ast_to_dict(ast, fold_ops=True)
        self.assertEqual(result['eight'], 8)
        self.assertEqual(result['eightyeight'], 88)
        self.assertEqual(result['wot'], 0)
        self.assertEqual(result['objobj'], 0)
        self.assertEqual(result['arrarr'], 0)
        self.assertEqual(result['six'], 6)
        self.assertEqual(result['sixteen'], 16)
        self.assertEqual(result['id'], 0)
        self.assertEqual(result['neg2'], 2147483646)
        self.assertEqual(result['negmax'], 2147483648)
        self.assertEqual(result['max'], 4294967295)
        self.assertEqual(result['zero'], 0)
        self.assertEqual(result['overflow'], 0)

    def test_binop_bitwise(self):
        ast = parse("""
        and_ = 12 & 10;
        xor_ = 12 ^ 10;
        or_ = 12 | 10;
        zero = 4294967296 & 4294967296;
        minus1 = 4294967295 & 4294967295;
        oneoneone = [111] & '111';
        nothing = {} ^ [];
        """)
        result = ast_to_dict(ast, fold_ops=True)
        self.assertEqual(result['and_'], 8)
        self.assertEqual(result['xor_'], 6)
        self.assertEqual(result['or_'], 14)
        self.assertEqual(result['zero'], 0)
        self.assertEqual(result['minus1'], -1)
        self.assertEqual(result['oneoneone'], 111)
        self.assertEqual(result['nothing'], 0)

    def test_binop_logical(self):
        ast = parse("""
        zero = ['asdf'] && 0 && 1;
        one = '1' && true && 1;
        two = '2' || 0 && 1;
        three = '33' && 0 || [3];
        wut = unknown && 1;
        none = unknown || null;
        """)
        result = ast_to_dict(ast, fold_ops=True)
        self.assertEqual(result['zero'], 0)
        self.assertEqual(result['one'], 1)
        self.assertEqual(result['two'], '2')
        self.assertEqual(result['three'], [3])
        # as the value bound to any identifier is assumed to be null,
        # the "variable" is returned as-is.
        self.assertEqual(result['wut'], 'unknown')
        self.assertEqual(result['none'], None)

    def test_binop_folding_various(self):
        ast = parse("""
        ten = 1 + 2 + 3 + 4;
        eleven = 1 + 2 * 3 + 4;
        lel = 'x' * 'x' + 'x'
        """)
        result = ast_to_dict(ast, fold_ops=True)
        self.assertEqual(result['ten'], 10)
        self.assertEqual(result['eleven'], 11)
        self.assertEqual(result['lel'], 'NaNx')

    def test_binop_folding_with_grouping(self):
        ast = parse("""
        twentysix = (1 + 2) * (3 + 4) + 5;
        check = ('x' * 'x') * ('x' * 'x');
        """)
        result = ast_to_dict(ast, fold_ops=True)
        self.assertEqual(result['twentysix'], 26)
        self.assertEqual(result['check'], 'NaN')

    def test_unary_expr_plusminus_folding(self):
        unparser = Unparser()
        ast = parse("""
        plus = +'42';
        neg = -'42';
        plusobj = +{};
        negobj = -{};
        """)
        self.assertEqual(dict(unparser(ast)), {
            'plus': 42,
            'neg': -42,
            'plusobj': 'NaN',
            'negobj': 'NaN',
        })

    def test_unary_expr_bitwise_not_folding(self):
        ast = parse("""
        nu = ~null;
        nn = ~NaN;
        neg1 = ~0;
        zero = ~'-1';
        negthreethree = ~[32];
        negtwofivesix = ~'0xff';
        bigzero = ~4294967295;
        smallneg1 = ~{};
        max31 = ~2147483648;
        min31 = ~2147483647;
        """)
        self.assertEqual(ast_to_dict(ast, fold_ops=True), {
            'nn': -1,
            'nu': -1,
            'neg1': -1,
            'zero': 0,
            'negthreethree': -33,
            'negtwofivesix': -256,
            'bigzero': 0,
            'smallneg1': -1,
            'max31': 2147483647,
            'min31': -2147483648,
        })

    def test_unary_expr_logical_not_folding(self):
        ast = parse("""
        nu = !null;
        nn = !NaN;
        t = !0;
        f = !1;
        tt = !false;
        ff = ![0];
        ttt = !'';
        fff = !{};
        """)
        self.assertEqual(ast_to_dict(ast, fold_ops=True), {
            'nn': True,
            'nu': True,
            't': True,
            'f': False,
            'tt': True,
            'ff': False,
            'ttt': True,
            'fff': False,
        })

    def test_unary_expr_various_folding(self):
        ast = parse("""
        de = delete obj;
        i = +1;
        j = -1;
        """)
        self.assertEqual(ast_to_dict(ast, fold_ops=True), {
            'de': 'delete obj',
            'i': 1,
            'j': -1,
        })
